/*
 * The project kernsh is (C) 2002-2004. This program is free software;
 * you can redistribute it and/or modify it under the terms of the GNU
 * General Public License as published by the Free Software Foundation;
 * Version 2. This guarantees your right to use, modify, and
 * redistribute this software under certain conditions.
 *
 * Source is provided to this software because we believe users have a
 * right to know exactly what a program is going to do before they run
 * it.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details (
 * http://www.gnu.org/copyleft/gpl.html ).
 *
 */

/*
 * Dictracy Loadable Kernel Module
 *                          by -  twiz   - twiz@email.it
 *                                thefly - thefly@acaro.org
 *
 * That module let you investigate, read, search, dump and write the virtual
 * address space of a process running.
 *
 * From the idea exposed by vecna in rmfbd :
 *    http://www.s0ftpj.org/bfi/dev/BFi11-dev-06
 *
 * Thanks : silvio, Liv Tyler (by thefly)
 */

#define MODULE
#define __KERNEL__

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif
#include <asm/uaccess.h>
#include <linux/sys.h>
#include <linux/smp_lock.h>
#include <linux/unistd.h>
#include <linux/pagemap.h> //include highmem in 2.4
#include <linux/file.h>
#include <asm/pgtable.h>
#include <nrsys.h> /* Automatically build */

#define __NR_getexeimage NR_sysfree0
#define __NR_getvirtaddr NR_sysfree1
#define __NR_rwfromvirt NR_sysfree2

#if defined( pid_hashfn )
struct
task_struct * get_task( pid_t pid )
{
    struct task_struct * p = NULL; //prevent bugs !!
    struct task_struct ** htable = &pidhash[ pid_hashfn(pid) ];

    read_lock( &tasklist_lock );
    for ( p = *htable; p && p->pid != pid; p = p->pidhash_next );
    read_unlock( &tasklist_lock );

    return( p );
}
#else
#warning : If You re using a Fedora Core kernel, please be carefull with this module !!
/* Notice : This is a temporary solution because it could be hijacked since 
 * it's an exported function (at least on my fedora core 1 kernel)
 */
struct
task_struct * get_task( pid_t pid )
{
	   return find_task_by_pid( pid );
}
#endif
/* modified from carbonite / author: foundstone */
int
getexeimage( int pid, char * filename )
{
    struct task_struct * task;
	struct vm_area_struct * vma;
	struct mm_struct * mm;
	struct file * f;
	unsigned char c;
	ssize_t ssize;
	mm_segment_t fs;
	int i;

	task = get_task( pid );
	if ( task == NULL ) return( 1 );
	if ( task->mm == NULL ) return( 1 );
	
	f = filp_open( filename, O_CREAT, S_IRWXU | S_IRWXG | S_IRWXO );
#if LINUX_VERSION_CODE >= 0x020400
	atomic_inc( &f->f_count );
#else
	f->f_count++;
#endif
	fs = get_fs();
	set_fs( KERNEL_DS );

	mm = task->mm;
	if ( mm ) {
		vma = mm->mmap;
		for ( vma = mm->mmap; vma; vma = vma->vm_next ) {
			if ( (vma->vm_flags & VM_EXECUTABLE) 
					&& (vma->vm_flags & VM_EXEC ) /* Avoid 2 dumps */
					&& vma->vm_file ) {
				vma->vm_file->f_pos = 0;
				ssize = vma->vm_file->f_op->read( vma->vm_file, &c, sizeof(c), 
					&vma->vm_file->f_pos );
				if ( ssize > 0 ) {
					while ( ssize != 0 ) {
						i = f->f_op->write( f, &c, sizeof(c), &f->f_pos );
						if ( i != ssize ) {
							filp_close( f, 0 );
							return( -1 );
						}	
						ssize = vma->vm_file->f_op->read( vma->vm_file, &c, sizeof(c), 
							&vma->vm_file->f_pos );
					}
				}
			}
		}
	}

	set_fs( fs );
	fput( f );
	filp_close( f, 0 );
	return( 0 );
}

struct mem_addr {
        unsigned long vm_start;
        unsigned long vm_end;
        unsigned long vm_start_cache;
        unsigned long vm_end_cache;
        int map_count;
        unsigned long startcode;
        unsigned long endcode;
        unsigned long startdata;
        unsigned long enddata;
        unsigned long startbrk;
        unsigned long startstack;
        unsigned long arg_start;
        unsigned long arg_end;
        unsigned long env_start;
        unsigned long env_end;
        unsigned long rss;
        unsigned long total_vm;
        unsigned long locked_vm;
        unsigned long swap_address;
};

int (*handle_mm_fault_e)( struct mm_struct *, struct vm_area_struct *, 
	unsigned long , int );
struct vm_area_struct *(*find_extend_vma_e)( struct mm_struct *, 
	unsigned long );
extern void* sys_call_table[];

int 
getvirtaddr( struct mem_addr * temmem, int pid )
{ 
	struct task_struct * temp;
	temp = get_task( pid );
	if ( temp == NULL ) {
		return( 1 );
  	}
	if ( temp->mm == NULL ) {
		return( 1 );
	}
	if ( temp->mm->mmap != NULL ) {	
		temmem->vm_start = temp->mm->mmap->vm_start;
		temmem->vm_end = temp->mm->mmap->vm_end;
	} else {
		temmem->vm_start = 0;
		temmem->vm_end = 0;
	}
	if ( temp->mm->mmap_cache != NULL ) {
		temmem->vm_start_cache = temp->mm->mmap_cache->vm_start;
		temmem->vm_end_cache = temp->mm->mmap_cache->vm_end;
	} else {
		temmem->vm_start_cache = 0;
		temmem->vm_end_cache =0;
	}
	temmem->map_count = temp->mm->map_count;
	temmem->startcode = temp->mm->start_code;
	temmem->endcode = temp->mm->end_code;
	temmem->startdata = temp->mm->start_data; /* Always null in ELF for a 2.2*/
	temmem->enddata = temp->mm->end_data;
	temmem->startbrk = temp->mm->start_brk;
	temmem->startstack = temp->mm->start_stack;
	temmem->arg_start = temp->mm->arg_start;
	temmem->arg_end = temp->mm->arg_end;
	temmem->env_start = temp->mm->env_start;
	temmem->env_end = temp->mm->env_end;
	temmem->rss = temp->mm->rss;
	temmem->total_vm = temp->mm->total_vm;
	temmem->locked_vm = temp->mm->locked_vm;
/* XXX I need to find an easy way to test for the existance of this variable
 * Perhaps a test with gcc in a makefile $(shell gcc -o $(BUILD)/test.o test.c
 * with test.c containing the previous variable ? [btw not very efficient]
 */
#ifndef _RH_
	temmem->swap_address = temp->mm->swap_address;
#endif
	return( 0 );
}

struct page * 
getpagefromaddr( unsigned long addr, struct task_struct * task ) 
{ 
	pgd_t * pgd;
	pmd_t * pmd;
	pte_t * pte;
	struct vm_area_struct * vma;
	struct mm_struct * mm;
 
startpoint:
#if LINUX_VERSION_CODE >= 0x020400
	task_lock( task );
#endif
	mm = task->mm;
#if LINUX_VERSION_CODE >= 0x020400
	if ( mm ) atomic_inc( &mm->mm_users );
	task_unlock( task );
#endif
	if ( !mm ) return( NULL );
#if LINUX_VERSION_CODE >= 0x020400
	spin_lock( &task->mm->page_table_lock );
#endif  
	/* pgd */
	pgd = pgd_offset( mm, addr );
	if ( pgd_none(*pgd) ) {
		goto fault;
	}
	if ( pgd_bad(*pgd) ) {
#if LINUX_VERSION_CODE >= 0x020400
		pgd_ERROR( *pgd );
#else
		printk("ptrace: bad page directory %08lx\n", pgd_val(*pgd));
		pgd_clear(pgd);
#endif
		goto error;
	}

	/* pmd */
	pmd = pmd_offset( pgd, addr );
	if ( pmd_none(*pmd) ) {
		goto fault;
	}
	if ( pmd_bad(*pmd) ) {
#if LINUX_VERSION_CODE >= 0x020400
		pmd_ERROR( *pmd );
#else
		printk("ptrace: bad page middle %08lx\n", pmd_val(*pmd));
		pmd_clear(pmd);
#endif
		goto error;
	} 
 
	/* pte */
#if !defined(pte_offset_kernel)
	pte = pte_offset( pmd, addr );
#else
	pte = pte_offset_kernel( pmd, addr ); /* ?? */
#endif
	if ( !pte_present(*pte) ) {
		goto fault;
	}

	/* Does Nothing :) */
	spin_unlock( &task->mm->page_table_lock );
	return( (struct page *)pte_page(*pte) );

fault:
	/*****************************************
	 * I think that you should pray to never 
	 * use this code :). I guess that Oopses
	 * will be the only thing that you get.
	 * - Kstat
	 */
	spin_unlock( &task->mm->page_table_lock );
	vma = find_extend_vma_e( mm, addr );
#if LINUX_VERSION_CODE >= 0x020400
	atomic_dec( &mm->mm_users );
#endif
	if ( handle_mm_fault_e(mm, vma, addr, 0) > 0 ) { 
		goto startpoint;
	} else {
		return( NULL );
	}
error:
	spin_unlock( &task->mm->page_table_lock );
#if LINUX_VERSION_CODE >= 0x020400
	atomic_dec( &mm->mm_users );
#endif
	return( NULL ); 
} 

int 
rwfromvirt( int pid, char * buffer, int len, unsigned long addr, int w )
{ 
	struct task_struct * task;
	struct page * page;
#if LINUX_VERSION_CODE >= 0x020400
	struct mm_struct * mm;
#endif
	char * maddr;
	
	if ( addr == 0 ) {
		return( -1 );
	}

	task = get_task( pid );
	if ( task == NULL ) return( 1 );
	page = getpagefromaddr( addr, task );
	if ( page == NULL ) return( -1 );
#if LINUX_VERSION_CODE >= 0x020400
	if ( w ) {
		if ( (!VALID_PAGE(page)) || PageReserved(page) ) return( -2 );
		maddr = (char*)kmap( page );
		memcpy( maddr + (addr & ~PAGE_MASK), buffer, len );
		flush_page_to_ram( page );
		flush_icache_page( find_extend_vma_e(task->mm, addr), page );
		kunmap( page );
	} else {
		maddr = (char *)kmap(page);
		memcpy(buffer, maddr + (addr & ~PAGE_MASK), len);
		flush_page_to_ram(page);
		kunmap(page);
	}
#else
	/*************************
	 * Be VERY Carefull with this. I do some basics tests 
	 * and have NO IDEA if this is really working !!! at least
	 * You should not get Oops ;)
	 * - Kstat
	 */
	if (MAP_NR(page) >= max_mapnr) return ( -2 );
	if ( w ) {
		if ( PageReserved(page) ) return ( -2 );
		maddr = (char *)((unsigned long int)page
				+ (addr & ~PAGE_MASK));
		memcpy(maddr,buffer,len);
		flush_page_to_ram( page );
	} else {
		maddr = (char *)((unsigned long int)page
				+ (addr & ~PAGE_MASK));
		memcpy(buffer,maddr,len);
		flush_page_to_ram( page );
	}
#endif
#if LINUX_VERSION_CODE >= 0x020400
	mm = task->mm;
	atomic_dec( &mm->mm_users );
#endif
	return( 0 );
}

static void * old_get = NULL;
static void * old_rw = NULL;
static void * old_exe = NULL;

int 
init_module() 
{ 
	old_get = sys_call_table[ __NR_getvirtaddr ];
	sys_call_table[ __NR_getvirtaddr ] = getvirtaddr;
	old_rw = sys_call_table[ __NR_rwfromvirt ];
	sys_call_table[ __NR_rwfromvirt ] = rwfromvirt;
	old_exe = sys_call_table[ __NR_getexeimage ];
	sys_call_table[ __NR_getexeimage ] = getexeimage;
	return( 0 );
} 

void 
cleanup_module()
{
	sys_call_table[ __NR_getvirtaddr ] = old_get;
	sys_call_table[ __NR_rwfromvirt ] = old_rw;
	sys_call_table[ __NR_getexeimage ] = old_exe;
} 

MODULE_LICENSE( "GPL" );
