/*
 * The project kernsh is (C) 2002-2004. This program is free software;
 * you can redistribute it and/or modify it under the terms of the GNU
 * General Public License as published by the Free Software Foundation;
 * Version 2. This guarantees your right to use, modify, and
 * redistribute this software under certain conditions.
 *
 * Source is provided to this software because we believe users have a
 * right to know exactly what a program is going to do before they run
 * it.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details (
 * http://www.gnu.org/copyleft/gpl.html ).
 *
 */

/* Be extremly Carrefull when playing with this file ...
 * Why ?
 * Because all the headers here comes from kernel space, not standard lib ...
 * Hope you know what your doing :) (me not :D) - Kstat
 */

#define __KERNEL__
#ifdef _LINUX_LIST_H
#undef _LINUX_LIST_H
#endif
#include <linux/list.h>
/* #include <linux/config.h> */
#ifdef GCC33
#define __ASM_SYSTEM_H
#define local_irq_save(x) do { x=x; } while (0);
#define local_irq_enable(x) do { } while (0);
#define local_irq_disable(x) do { } while (0);
#define local_irq_restore(x) do { } while (0);
#define cmpxchg(x,y,z) 1;
#endif
#include <linux/mm.h>
/* #include <linux/string.h> */
#include <linux/dcache.h>
#include <linux/fs.h>
#include <linux/version.h>
#ifdef __LINUX_VMALLOC_H
#error already done !
#endif
#include <linux/vmalloc.h>
#undef __KERNEL__

/* #include <stdlib.h> */
#include <getopt.h>
#include "include/kernsh.h"

/* From stdlib.h */
extern unsigned long int strtoul (__const char *__restrict __nptr, 
	char **__restrict __endptr, int __base)
__THROW;

extern void free (void *ptr);
/* Temporary location for displaying vm_struct system */
#ifndef VM_ALLOC
#define VM_ALLOC 0x00000002
#endif
void
display_vm_struct( struct vm_struct vm )
{
	fprintf ( stdout,
		"\n"
		"  Flags : %s (%08x)\n"
		"Address : %p\n"
		"   Size : %lx or %lu\n"
		"   Next : %p\n",
		(vm.flags & VM_ALLOC) ? "VM_ALLOC" : "VM_IOREMAP", vm.flags,
		vm.addr,
		vm.size, vm.size,
		vm.next
	);
	return;
}

void
check_addr(unsigned int n, unsigned long table[])
{
	int i = 0,j;
	unsigned long test;

	for ( i = 0; i < n;i++ ) {
		test = table[i];
		if ( test == 0L ) {
			fprintf( stdout, "Strange Value : %p\n", (void *)test );
			continue;
		}
		for ( j = i + 1; j < n; j++ ) {
			if ( test == table[j] ) {
				fprintf( stdout, "Values are equals ! : %p\n",(void *)test);
			}
		}
	}
}

void
follow_vm_list( void )
{
	struct vm_struct svl, * first;
	unsigned long table_vmalloc[1024];
	size_t vmlist = get_symbol_by_name( "vmlist" );
	unsigned int pos_table_vm = 0;
	int n = 0;
	
	if ( vmlist == 0 ) {
		fatal( "Can't find the vmlist !\n" );
		return;
	}

	n = read_kmem( &first, vmlist, sizeof(first), file_descriptor_kmem );
	if ( n < 0 ) {
		fatal( "Can't read vmlist !\n" );
		return;
	}

	do {
		n = read_kmem( &svl, (size_t)first, sizeof(svl), file_descriptor_kmem );
		if ( n < 0 ) {
			return;
		}
		
		table_vmalloc[pos_table_vm++] = (unsigned long)svl.addr;
		display_vm_struct( svl );
		check_addr(pos_table_vm,table_vmalloc);

		first = svl.next;
		if ( svl.flags & VM_ALLOC ) {
			get_info_module( NULL, (size_t)svl.addr );
			getchar();
		}
	} while ( first != NULL );
	/* Check if all svl.addr are right */
	check_addr(pos_table_vm,table_vmalloc);
	return;
}

/* This function do basic test against all {file/inode/dentry/super}_operations
 * symbol found in System.map (missing is block_device)
 *
 * Better to make a test with and without mmap_read !!
 */
#define TEST_IN_KERNEL( op, sfops ) \
	if ( fops.op != NULL && ((unsigned int)fops.op < PAGE_OFFSET || (unsigned int)fops.op > PAGE_OFFSET + 0x00300000) ) { \
		error++; \
		fprintf( stdout, "  This operation (%s) seems not good : %p\n", sfops, fops.op ); \
	}

#if ! defined(KERNEL_VERSION)
#error KERNEL_VERSION is not defined !
#endif

int
check_struct_fops( struct file_operations fops )
{
	/* XXX Must be tested against a 2.2 Kernel !!! */
	int error = 0;
	/* This is very linear basic checks see linux/fs.h */
	TEST_IN_KERNEL(llseek,"llseek");
	TEST_IN_KERNEL(read,"read");
	TEST_IN_KERNEL(write,"write");
	TEST_IN_KERNEL(readdir,"readdir");
	TEST_IN_KERNEL(poll,"poll");
	TEST_IN_KERNEL(ioctl,"ioctl");
	TEST_IN_KERNEL(mmap,"mmap");
	TEST_IN_KERNEL(open,"open");
	TEST_IN_KERNEL(flush,"flush"); /* in 2.4 */
	TEST_IN_KERNEL(release,"release");
	TEST_IN_KERNEL(fsync,"fsync");
	TEST_IN_KERNEL(fasync,"fasync");
#if LINUX_VERSION_CODE < 0x020400
	TEST_IN_KERNEL(check_media_change,"check_media_change"); /* in 2.2 ? */
	TEST_IN_KERNEL(revalidate,"revalidate"); /* in 2.2 ? */
#endif
	TEST_IN_KERNEL(lock,"lock");
#if LINUX_VERSION_CODE >= 0x020400
	TEST_IN_KERNEL(readv,"readv"); /* in 2.4 */
	TEST_IN_KERNEL(writev,"writev"); /* in 2.4 */
	TEST_IN_KERNEL(sendpage,"sendpage"); /* in 2.4 */
	TEST_IN_KERNEL(get_unmapped_area,"get_unmapped_area"); /* in 2.4 */
#endif
	return( error );
}

int
check_struct_iops( struct inode_operations fops )
{
	int error = 0;
	/*error = check_my_file_operation ( fops.default_file_ops );*//* see in 2.2 ?*/
	TEST_IN_KERNEL(create,"create");
	TEST_IN_KERNEL(lookup,"lookup");
	TEST_IN_KERNEL(link,"link");
	TEST_IN_KERNEL(unlink,"unlink");
	TEST_IN_KERNEL(symlink,"symlink");
	TEST_IN_KERNEL(mkdir,"mkdir");
	TEST_IN_KERNEL(rmdir,"rmdir");
	TEST_IN_KERNEL(mknod,"mknod");
	TEST_IN_KERNEL(rename,"rename");
	TEST_IN_KERNEL(readlink,"readlink");
	TEST_IN_KERNEL(follow_link,"follow_link");
#if LINUX_VERSION_CODE < 0x020400
	TEST_IN_KERNEL(readpage,"readpage"); /* 2.2 ? */
	TEST_IN_KERNEL(writepage,"writepage"); /* 2.2 ? */
	TEST_IN_KERNEL(bmap,"bmap"); /* 2.2 */
#endif
	TEST_IN_KERNEL(truncate,"truncate");
	TEST_IN_KERNEL(permission,"permission");
#if LINUX_VERSION_CODE < 0x020400
	TEST_IN_KERNEL(smap,"smap"); /* 2.2 ? */
	TEST_IN_KERNEL(updatepage,"updatepage"); /* 2.2 ? */
#endif
	TEST_IN_KERNEL(revalidate,"revalidate");
#if LINUX_VERSION_CODE < 0x020400
	TEST_IN_KERNEL(prepare_write,"prepare_write");
	TEST_IN_KERNEL(sync_page,"sync_page");
#endif
#if LINUX_VERSION_CODE >= 0x020400
	TEST_IN_KERNEL(setattr,"setattr"); /* only in 2.4 ? */
	TEST_IN_KERNEL(getattr,"getattr"); /* only in 2.4 ? */
#endif
	/* TEST_IN_KERNEL(setxattr,"setxattr"); *//* only in 2.4 ? */
	/* TEST_IN_KERNEL(getxattr,"getxattr"); *//* only in 2.4 ? */
	/* TEST_IN_KERNEL(listxattr,"listxattr"); *//* only in 2.4 ? */
	/* TEST_IN_KERNEL(removexattr,"removexattr"); *//* only in 2.4 ? */

	return( error );
}

int
check_struct_dops( struct dentry_operations fops )
{
	int error =  0;
	TEST_IN_KERNEL(d_revalidate,"d_revalidate");
	TEST_IN_KERNEL(d_hash,"d_hash");
	TEST_IN_KERNEL(d_compare,"d_compare");
	TEST_IN_KERNEL(d_delete,"d_delete");
	TEST_IN_KERNEL(d_release,"d_release");
	TEST_IN_KERNEL(d_iput,"d_iput");

	return( error );
}

int
check_struct_sops( struct super_operations fops )
{
	int error = 0;
	/* TEST_IN_KERNEL(alloc_inode,"alloc_inode"); *//* in 2.4 */
	/* TEST_IN_KERNEL(destroy_inode,"destroy_inode"); *//* in 2.4 */
	TEST_IN_KERNEL(read_inode,"read_inode");
#if LINUX_VERSION_CODE >= 0x020400
	TEST_IN_KERNEL(read_inode2,"read_inode2"); /* in 2.4 and reiserfs ?*/
	TEST_IN_KERNEL(dirty_inode,"dirty_inode"); /* in 2.4 */
#endif
	TEST_IN_KERNEL(write_inode,"write_inode");
	TEST_IN_KERNEL(put_inode,"put_inode");
	TEST_IN_KERNEL(delete_inode,"delete_inode");
	TEST_IN_KERNEL(put_super,"put_super"); /* in 2.4 */
	TEST_IN_KERNEL(write_super,"write_super"); /* in 2.4 */
	/* TEST_IN_KERNEL(notify_change,"notify_change");*/ /* in 2.2 ? */
	/* TEST_IN_KERNEL(sync_fs,"sync_fs"); *//* in 2.4 */
#if LINUX_VERSION_CODE >= 0x020400
	TEST_IN_KERNEL(write_super_lockfs,"write_super_lockfs"); /* in 2.4 */
	TEST_IN_KERNEL(unlockfs,"unlockfs"); /*in 2.4*/
#endif
	TEST_IN_KERNEL(statfs,"statfs");
	TEST_IN_KERNEL(remount_fs,"remount_fs");
	TEST_IN_KERNEL(clear_inode,"clear_inode");
	TEST_IN_KERNEL(umount_begin,"umount_begin"); /* in 2.4 */
#if LINUX_VERSION_CODE >= 0x020400
	TEST_IN_KERNEL(fh_to_dentry,"fh_to_dentry"); /* in 2.4 and knfd ? */
	TEST_IN_KERNEL(dentry_to_fh,"dentry_to_fh"); /* in 2.4 and knfsd ? */
	TEST_IN_KERNEL(show_options,"show_options"); /* in 2.4 and knfsd ? */
#endif
	return( error );
}

int
check_my_file_operation( size_t address )
{
	struct file_operations fops;
	int n;

	n = read_kmem( &fops, address, sizeof(fops), file_descriptor_kmem );
	if ( n < 0 ) {
		perror( "check_my_file_operation:read_kmem()" );
		return( -1 );
	}

	/* fprintf(stdout, " Checking %#08x \n",address);*/
	n = check_struct_fops( fops );
	if ( n < 0 ) {
		fatal ( "Can't check this file operation 0x%08lx!\n", address );
	} else  if ( n > 0 ) {
		char buf[ 128 ];
		char * tmp = NULL;
		snprintf( buf, sizeof(buf), "%#08x", address );
		tmp = get_symbol_by_addr(buf);
		fprintf( stdout, "This fops seems compromized ! %s \n", tmp );
		/* free get_symbol_by_addr */
		free(tmp);
		tmp = NULL;
	}

	return( n );
}

int
check_my_inode_operation( size_t address )
{
	struct inode_operations iops;
	int n = 0;

	n = read_kmem( &iops, address, sizeof(iops), file_descriptor_kmem );
	if ( n < 0 ) {
		perror( "check_my_inode_operation:read_kmem()" );
		return( -1 );
	}

	n = check_struct_iops( iops );
	if ( n < 0 ) {
		fatal ( "Can't check this Inode operation 0x%08lx!\n", address );
	} else if ( n > 0 ) {
		char buf[ 128 ];
		char * tmp = NULL;
		snprintf( buf, sizeof(buf), "%#08x", address );
		tmp = get_symbol_by_addr(buf);
		fprintf( stdout, "This iops seems compromized ! %s \n", tmp );
		/* free get_symbol_by_addr */
		free(tmp);
		tmp = NULL;
	}
	return( n );
}

int
check_my_dentry_operation( size_t address )
{
	struct dentry_operations dops;
	int n = 0;

	n = read_kmem( &dops, address, sizeof(dops), file_descriptor_kmem );
	if ( n < 0 ) {
		perror( "check_my_inode_operation:read_kmem()" );
		return( -1 );
	}

	n = check_struct_dops( dops );
	if ( n < 0 ) {
		fatal ( "Can't check this Inode operation 0x%08lx!\n", address );
	} else if ( n > 0 ) {
		char buf[ 128 ];
		char * tmp = NULL;
		snprintf( buf, sizeof(buf), "%#08x", address );
		tmp = get_symbol_by_addr(buf);
		fprintf( stdout, "This iops seems compromized ! %s \n", tmp );
		/* free get_symbol_by_addr */
		free(tmp);
		tmp = NULL;
	}
	return( n );
}

int
check_my_super_operation( size_t address )
{
	struct super_operations sops;
	int n = 0;

	n = read_kmem( &sops, address, sizeof(sops), file_descriptor_kmem );
	if ( n < 0 ) {
		perror( "check_my_super_operation:read_kmem()" );
		return( -1 );
	}

	n = check_struct_sops( sops );
	if ( n < 0 ) {
		fatal ( "Can't check this Super operation 0x%08lx!\n", address );
	} else if ( n > 0 ) {
		char buf[ 128 ];
		char * tmp = NULL;
		snprintf( buf, sizeof(buf), "%#08x", address );
		tmp = get_symbol_by_addr(buf);
		fprintf( stdout, "This sops seems compromized ! %s \n", tmp );
		/* free get_symbol_by_addr */
		free(tmp);
		tmp = NULL;
	}
	return( n );
}

int
check_my_file_operation_by_name( char * sfops )
{
	size_t address = get_symbol_by_name( sfops );

	if ( address != 0 ) {
		return( check_my_file_operation(address) );
	} else {
		fatal( "Symbol not found (%s) !\n", sfops );
		return( -1 );
	}
}

int
check_my_inode_operation_by_name( char * siops )
{
	size_t address = get_symbol_by_name( siops );

	if ( address != 0 ) {
		return( check_my_inode_operation(address) );
	} else {
		fatal( "Symbol not found (%s) !\n", siops );
		return( -1 );
	}
}

int
check_my_dentry_operation_by_name( char * sdops )
{
	size_t address = get_symbol_by_name( sdops );

	if ( address != 0 ) {
		return( check_my_dentry_operation(address) );
	} else {
		fatal( "Symbol not found (%s) !\n", sdops );
		return( -1 );
	}
}

int
check_my_super_operation_by_name( char * ssops )
{
	size_t address = get_symbol_by_name( ssops );

	if ( address != 0 ) {
		return( check_my_super_operation(address) );
	} else {
		fatal( "Symbol not found (%s) !\n", ssops );
		return( -1 );
	}
}

void
check_file_operation ( void )
{
	size_t address = 2;
	int n;
	int pwet = 0;

	fprintf( stdout, "Beginning checks ...\n" );
	while ( (address = get_symbol_by_system_map("_file_operation", 
		address)) != 0 ) {
		n = check_my_file_operation( address );
		if ( n > 0 ) {
			pwet ++;
		}
	}

	address = 2;
	while ( (address = get_symbol_by_system_map ("_fops",address )) != 0 ) {
		n = check_my_file_operation ( address );
		if ( n > 0 ) {
			pwet ++;
		}
	}

	if ( pwet != 0 ) {
		fatal( "WARNING it seems that you have a problem !\n" );
	} else {
		fprintf( stdout, "All test ok !\n" );
	}
	return;
}

void
check_inode_operation( void )
{
	size_t address = 2;
	int n;
	int pwet = 0;

	fprintf( stdout, "Beginning checks ...\n" );

	while ( (address = get_symbol_by_system_map("_inode_operation", 
		address )) != 0 ) {
		n = check_my_inode_operation ( address );
		if ( n > 0 ) {
			pwet ++;
		}
	}

	address = 2;
	while ( (address = get_symbol_by_system_map("_iops",address )) != 0 ) {
		n = check_my_inode_operation ( address );
		if ( n > 0 ) {
			pwet ++;
		}
	}
    
	if ( pwet != 0 ) {
		fatal( "WARNING it seems that you have a problem !\n" );
	} else {
		fprintf( stdout, "All test ok !\n" );
	}

	return;
}

void
check_dentry_operation( void )
{
	size_t address = 2;
	int n;
	int pwet = 0;

	fprintf( stdout, "Beginning checks ...\n" );
	while ( (address = get_symbol_by_system_map("_dentry_operation", 
		address)) != 0 ) {
		n = check_my_dentry_operation( address );
		if ( n > 0 ) {
			pwet ++;
		}
	}

	address = 2;
	while ( (address = get_symbol_by_system_map("_dops", address)) != 0 ) {
		n = check_my_dentry_operation( address );
		if ( n > 0 ) {
			pwet ++;
		}
	}

	if ( pwet != 0 ) {
		fatal( "WARNING it seems that you have a problem !\n" );
	} else {
		fprintf( stdout, "All test ok !\n" );
	}
	return;
}

void
check_super_operation( void )
{
	size_t address = 2;
	int n = 0;
	int pwet = 0;

	while ( (address = get_symbol_by_system_map("_sops",address )) != 0 ) {
		n = check_my_super_operation( address );
		if ( n > 0 ) {
			pwet ++;
		}
	}

	if ( pwet != 0 ) {
		fatal( "WARNING it seems that you have a problem !\n" );
	} else {
		fprintf( stdout, "All test ok !\n" );
	}
	return;
}

void
check_functions_hook ( void )
{
	/* we need to check for "\xB8\x00\x00\x00\x00\xFF\xE0" into memory */
	size_t mm = (size_t)PAGE_OFFSET;
	int p = 0;
	char  buf[sizeof(size_t) * 2 + 3 /* 0x */ + 1/* \0 */ ];
	char * symbol = NULL;
	size_t to = 0;
	int print = 0;
	
	char pattern[] = "\xB8\x00\x00\x00\x00\xFF\xE0";
	
	fprintf(stdout,"Checking from %#08x to %#08x\n",mm,get_symbol_by_name("high_memory"));

	do {
		mm = kfind_area(mm,get_symbol_by_name("high_memory"),
				pattern,sizeof(pattern)-1,0);
		if ( mm != 0 ) {
			p++;
			fprintf(stdout,"Found a possible hook at %#08x to ",mm );
			if( read_kmem(&to,mm+1,sizeof(to),file_descriptor_kmem) < 0 ) {
				fatal("check_functions_hook : Can't read the address !\n");
				to = 0;
			}
			fprintf(stdout,"%#08x in (",to);
			snprintf(buf,sizeof(buf)-1,"%#08x",mm - 1); /* no frame */
			symbol = get_symbol_by_addr ( buf ) ;
			if ( symbol != NULL && 
					!(symbol[0] == '0' && symbol[1] == 'x') ) {
				fprintf(stdout,"%s", symbol );
				print++;
			}
			free(symbol);
			symbol = NULL;
			if ( !print ) {
				snprintf(buf,sizeof(buf)-1,"%#08x",mm - 3); /* with frame */
				symbol = get_symbol_by_addr ( buf ) ;
				if ( symbol != NULL &&
						!(symbol[0] == '0' && symbol[1] == 'x') ) {
					fprintf(stdout,"%s", symbol );
					print++;
				}
				free(symbol);
				symbol = NULL;
			}
			if ( !print ) {
			/* This must be the LAST !!! */
				snprintf(buf,sizeof(buf)-1,"%#08x",mm ); /* directly :) */
				symbol = get_symbol_by_addr ( buf ) ;
				if ( symbol != NULL) {
						fprintf(stdout,"%s", symbol );
						print++;
				}
				free(symbol);
				symbol = NULL;
			}
			if ( print ) {
				fprintf(stdout,")\n");
			}
			print = 0;
			mm += sizeof(pattern)-1;
		}
	} while ( mm != 0 );

	if ( p == 0 ) {
		fprintf(stdout, "Your system seems clean\n");
	} else if ( p == 1 ) {
		fprintf(stdout, "Your system looks Ok, I found one indirect jmp\n"
				" But if it is near 0xc0100000, it's the first kernel jmp\n"
			   );
	} else {
		fprintf(stdout, "WARNING: I found %d hooks on your system !\n",p);
	}
}
			
void
usage_check( void )
{
    usage_t p_usage;

    p_usage.command_line =
		"check [ -fgacijbderstyz ] <arguments> [WARNING TESTED ONLY on 2.4]";
    p_usage.description =
		"This command give you the possibility to checks some kernel\n"
		"structures that are modified by some rootkit. DON'T FORGET to\n"
		"run them with different reading technic (see kmem -ul)!";
    p_usage.arguments =
        "  -a : Check all file operations found in System.map (some could\n"
        "be forgotten and False Positive are possible) ;\n"
        "  -f [ address ] : check the file_operation structure at this "
        "address ;\n"
        "  -g [ name ] : check the file_operation structure with this name ;\n"
        "  -c : Check all inode operations found in System.map (some could\n"
        "be forgotten and False Positive are possible) ;\n"
        "  -i [ address ] : check the inode_operation structure at this "
        "address ;\n"
        "  -j [ name ] : check the inode_operation structure with this name ;\n"
        "  -b : Check all dentry operations found in System.map (some could\n"
        "be forgotten and False Positive are possible) ;\n"
        "  -d [ address ] : check the dentry_operation structure at this "
        "address ;\n"
        "  -e [ name ] : check the dentry_operation structure with this "
        "name ;\n"
        "  -r : Check all super operations found in System.map (some could be\n"
        "forgotten and False Positive are possible) ;\n"
        "  -s [ address ] : check the super_operation structure at this address ;\n"
        "  -t [ name ] : check the super_operation structure with this name ;\n"
        "  -y : basic checks for LKH type hooks (only in kernel text) ;\n"
        "  -z : Undocumented feature :)";

    usage( &p_usage );
    return;
}

void
do_check ( int pargc, char * pargv[] )
{
	size_t l;
	char c;

	NOARGUMENTS( pargc, 1, check );
	optind = 0;

	while ( (c = getopt(pargc, pargv, "zyabcd:e:f:g:i:j:rs:t:")) != EOF ) {
		switch ( c ) {
		case 'y':
			check_functions_hook();
			break;
		case 'z':
			follow_vm_list();
			break;
		case 'f':
			l = strtoul( optarg, NULL, 0 );
			check_my_file_operation( l );
			break;
		case 'g':
			check_my_file_operation_by_name( optarg );
			break;
		case 'a':
			check_file_operation();
			break;
		case 'i':
			l = strtoul( optarg, NULL, 0 );
			check_my_inode_operation( l );
			break;
		case 'j':
			check_my_inode_operation_by_name( optarg );
			break;
		case 'c':
			check_inode_operation();
			break;
		case 'd':
			l = strtoul( optarg, NULL, 0 );
			check_my_dentry_operation( l );
			break;
		case 'e':
			check_my_dentry_operation_by_name( optarg );
			break;
		case 'b':
			check_dentry_operation( );
			break;
		case 's':
			l = strtoul( optarg, NULL, 0 );
			check_my_super_operation( l );
			break;
		case 't':
			check_my_super_operation_by_name( optarg );
			break;
		case 'r':
			check_super_operation( );
			break;
		default:
			usage_check();
			return;
		}
	}
	return;
}

