/*
 * The project kernsh is (C) 2002-2004. This program is free software;
 * you can redistribute it and/or modify it under the terms of the GNU
 * General Public License as published by the Free Software Foundation;
 * Version 2. This guarantees your right to use, modify, and
 * redistribute this software under certain conditions.
 *
 * Source is provided to this software because we believe users have a
 * right to know exactly what a program is going to do before they run
 * it.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details (
 * http://www.gnu.org/copyleft/gpl.html ).
 *
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <sys/mman.h>
#include <elf.h>
#include <string.h>
#include <stdlib.h>
#include "include/kernsh.h"
/*#undef LIBDISASM_NO_COMPAT*/
#include "../libdisasm/libdis.h"

/* Ask the syntax of debug output */
//static enum x86_asm_format format_dis = att_syntax;
int disasm_syn = (int)att_syntax;

/* Ripped from my own works so, under GPL V2 :) - Kstat */
/* Some macros */
#define DTYPE() i.destType
#define CODE() i

#ifndef LIBDISASM_NO_COMPAT 
int
get_jmp_off( struct instr i )
{
	if ( (DTYPE () & OP_TYPE_MASK) != (OP_REL) ) {
		return( 0 );
	}

	/* The operand is a byte  */
	if ( (DTYPE() & (OP_SIZE_MASK)) == OP_BYTE ) {
		if ( ((DTYPE() & (OP_MOD_MASK)) & (OP_SIGNED)) == (OP_SIGNED) ) {
			if ( strtol(CODE().dest, NULL, 16) + CODE().size > 127 ) {    /* 2^7 - 1 */
				return( strtol(CODE().dest, NULL, 16)+CODE().size-256 );
			} else {
				return( strtol(CODE().dest, NULL, 16)+CODE().size );
			}
		}
	}

	if ( (DTYPE() & (OP_SIZE_MASK)) == OP_HWORD ) {
		if ( ((DTYPE() & (OP_MOD_MASK)) & (OP_SIGNED)) == (OP_SIGNED) ) {
			if ( strtol(CODE().dest, NULL, 16) + CODE().size > 32767 ) {  /* 2^15 - 1 */
				return( strtol(CODE().dest, NULL, 16)+CODE().size-65536 );
			} else {
				return( strtol(CODE().dest, NULL, 16)+CODE().size );
			}
		}
	}

	if ( (DTYPE() & (OP_SIZE_MASK)) == OP_WORD ) {
		if ( ((DTYPE() & (OP_MOD_MASK)) & (OP_SIGNED)) == (OP_SIGNED) ) {
			if ( strtol(CODE().dest, NULL, 16) + CODE ().size > 2147483647 ) { /* 2^31 - 1 */
				return( strtol(CODE().dest, NULL, 0)+CODE().size-4294967296 );
			} else {
				return( strtol(CODE().dest, NULL, 16)+CODE().size );
			}
		}
	}

	/* Never see better :) */
	return( strtol(CODE().dest, NULL, 0)+CODE().size );
}
#else
unsigned long
get_jmp_off( x86_op_t * op, unsigned long addr )
{
	if ( op->type != op_relative ) {
		return( addr );
	}
	
	switch( op->datatype ) {
		case op_byte :
			if ( (op->flags & op_signed) ) {
				return addr + op->data.sbyte ;
			} else {
				return addr + op->data.byte ;
			}
			break;
		case op_word :
			if ( (op->flags & op_signed) ) {
				return addr + op->data.sword;
			} else {
				return addr + op->data.word;
			}
			break;
		case op_dword :
			if ( (op->flags & op_signed) ) {
				return addr + op->data.sdword;
			} else {
				return addr + op->data.dword;
			}
			break;
		case op_qword :
//			if ( (op->flags & op_signed) ) {
				return addr + op->data.sqword;
//			} else {
//				return addr + op->data.qword;
//			}
			break;
		default:
			fprintf(stdout,"Warning : unknow kind of return\n");
			break;
	}
	return( addr + op->data.dword ); // Hope :)
}
#endif

#ifdef LIBDISASM_NO_COMPAT
/* The error reporter for libdisasm */
/* this is x86dis.c code that needs to be ripped off and used in x86dis.c */
void 
kernsh_reporter( enum x86_report_codes code, void *arg ) 
{
    char * str;

    /* here we could examine the error and do something useful;
     * instead we just print that an error occurred */
    switch ( code ) {
        case report_disasm_bounds:
            str = "Attempt to disassemble RVA beyond end of buffer";
            break;
        case report_insn_bounds:
            str = "Instruction at RVA extends beyond buffer \n\t\t\t"
				"(increase the size of the buffer !)";
            break;
        case report_invalid_insn:
            str = "Invalid opcode at RVA";
            break;
        case report_unknown:
        default:    /* make GCC shut up */
            str = "Unknown Error";
            break;
    }

    fprintf(stderr, "DISASM ERROR \'%s:\' 0x%08X\n", str, (int)arg);
}
#endif

/* this is a Stack Rebuilder :) */
#define STACK() tmp = (struct Stack *)malloc( sizeof(struct Stack) ); \
	if ( tmp == NULL ) { perror("STACK:malloc()"); return( -1 ); } \
	tmp->prev = stack; stack = tmp;
#ifdef LIBDISASM_NO_COMPAT
struct Stack {
	unsigned char code[ 32 ];
	int size;
	struct Stack * prev;
};

typedef struct Stack stack_t;

/* Need to be rebuilded with the new API :) */

int
make_esp( unsigned char * buf, size_t length, unsigned char * rest,
		size_t * rest_size )
{
	int err = -1;
	int size = 0;
	int pos = 0;
	x86_insn_t code;
	stack_t * stack = NULL, *tmp = NULL;
	
	int st = 0;

	/* Initialisation of the disassembler */
	x86_init(opt_none,kernsh_reporter);
	
	while ( pos < length ) {
		size = x86_disasm(buf,length,0,pos, &code);
		if ( size == 0 ) {
			err = -1;
			goto exit;
		}
		/* Ok We need to test the beginning of the function */
		/* we want to rebuild the esp in the right way */
		/* The function that we derivate MUST be compiled with
		 * -fomit-frame-pointer !!!! */
		switch (code.group) {
			case insn_stack :
				if ( code.type == insn_push || code.type == insn_pop ) {
					STACK();
					stack->code[0] = code.bytes[0];
				   	stack->code[0] += (code.type == insn_push) ? 
						(char) 8 : (char) -8 ;
					st += stack->size = 1;
				} else {
					char line[1024];
					if ( x86_format_insn(&code,line,sizeof(line),
								native_syntax) ) {
						fprintf(stdout,"DEBUG stack: %s\n",line);
					} else {
						err = -1;
						goto exit;
					}
				}
				break;
			case insn_arithmetic :
				if ( ( code.type == insn_sub || code.type == insn_add )
						&& code.operands[op_dest].type == op_register
						&& code.operands[op_dest].data.reg.id 
							== x86_sp_reg() ) {
					switch( code.size ) {
						case 3:
							STACK();
							memcpy(stack->code,code.bytes,3);
							stack->code[1] = ( code.type == insn_add )
								? (char) 0xEC : (char)0xC4 ;
							st += stack->size=3;
							break;
						case 6:
							STACK();
							memcpy(stack->code,code.bytes,6);
							stack->code[1] = ( code.type == insn_add ) 
								? (char) 0xEC : (char)0xC4 ;
							st += stack->size=6;
							break;
						default:
						{
							char line[1024];
		                    if ( x86_format_insn(&code,line,sizeof(line),
										native_syntax) ) {
								fprintf(stdout,"DEBUG: %s\n",line);
				            } else {
								err = -1;
								goto exit;
							}
						}
						break;
                    }	
				} else if ( code.type == insn_dec || code.type == insn_inc ) {
					if ( code.operands[op_dest].type == op_register ) {
						STACK();
						stack->code[0] = code.bytes[0];
					   	stack->code[0] += (code.type == insn_dec) ? (char) -8 :
							(char) 8;
						st += stack->size = 1;
					} else {
						char line[1024];
						if ( x86_format_insn(&code,line,sizeof(line),
									native_syntax) ) {
							fprintf(stdout,"DEBUG: %s\n",line);
						} else {
							err = -1;
							goto exit;
						}
					}
				}
				break;
			case insn_controlflow :
				err = pos;
				goto revert;
			default :
			   {
					char line[1024];
	                if ( x86_format_insn(&code,line,sizeof(line),
		                         native_syntax) ) {
			            fprintf(stdout,"DEBUG: %s\n",line);
				    } else {
					    err = -1;
	                    goto exit;
		            }
				}
				break;
		}
		pos += size; /* Next instruction */
	}
	
	/* Revert the stack that we just create */
revert:
	if ( st <= *rest_size ) {
		*rest_size = 0;
		while ( stack ) {
			for ( size = 0; size < stack->size; size++ ) {
				rest[(*rest_size)++] = stack->code[size];
			}
			tmp = stack;
			stack = tmp->prev;
			free( tmp );
		}
	} else {
		fatal( "Buffer to short (need %d)!\n", pos );
		err = -1;
	}
	
	/* exit the disassembler */
exit:
	x86_cleanup();
	return err;
}
#endif
#ifndef LIBDISASM_NO_COMPAT
int
make_esp( unsigned char * buf, size_t length, unsigned char * rest, 
	size_t * rest_size )
{
	int err = 0;
	int size = 0;
	int pos = 0;
	struct instr code;
	struct Stack {
		int size;
		unsigned char code[ 12 ];
		struct Stack * prev;
	};
	int st = 0;
	struct Stack * stack, * tmp;

	stack = (struct Stack *) malloc( sizeof(struct Stack) );
	memset( stack, 0, sizeof(stack) );
	disassemble_init( 0, ATT_SYNTAX );

	while ( pos < length ) {
		/* perhaps need to remove length - pos ! */
		size = disassemble_address( (char *)buf+pos, length - pos , &code );
		if ( !size ) {
			err = -1;
			goto exit;
		}

		if ( (code.mnemType & (INS_TYPE_MASK)) == (INS_SUB) ) {
			if ( strncmp(code.dest, "%esp", 4) == 0 ) {
				/* we must reverse this */
				STACK ();
				/* so do an add */
				switch ( code.size ) {
				case 3:
					/* fprintf(stderr,"Code add 3\n"); */
					stack->code[0] = (char) 0x83;
					stack->code[1] = (char) 0xC4;
					stack->code[2] = buf[pos + 2];
					stack->size = 3;
					break;
				case 6:
					/* fprintf(stderr,"Code add 6\n"); */
					stack->code[0] = (char) 0x81;
					stack->code[1] = (char) 0xC4;
					stack->code[2] = buf[pos + 2];
					stack->code[3] = buf[pos + 3];
					stack->code[4] = buf[pos + 4];
					stack->code[5] = buf[pos + 5];
					stack->size = 6;
					break;
				default:
					err = -1;
					goto exit;
				}
				st = pos + code.size;
				/* break;*/
				/* stop to disassemble ? to be verified and tested */
			}
		} else if ( (code.mnemType & (INS_TYPE_MASK)) == (INS_PUSH) ) {
			STACK ();
			stack->code[0] = buf[pos] + (char) 8; /* convert a push into a pop */
			stack->size = 1;
		} else {
			/* The rest is not being take in account ? */
			if ( (code.mnemType & (INS_TYPE_MASK)) == (INS_CMP) ) {
				/* This is the value to erase !! */
				err = pos;
				break;
			} else if ( (code.mnemType & (INS_TYPE_MASK)) == (INS_DEC) ) {
				STACK ();
				stack->code[0] = buf[pos] - (char) 8; /* Convert a DEC to an INC */
				stack->size = 1;
			} else {
				/* Just in case there is no sub esp, if we found a jump/call gave up */
				if ( ((code.mnemType & (INS_TYPE_MASK)) == (INS_BRANCH))
					|| ((code.mnemType & (INS_TYPE_MASK)) == (INS_BRANCHCC))
					|| ((code.mnemType & (INS_TYPE_MASK)) == (INS_CALL))
					|| ((code.mnemType & (INS_TYPE_MASK)) == (INS_CALLCC))
					|| ((code.mnemType & (INS_TYPE_MASK)) == (INS_RET)) ) {
					fatal( "short exit %s !\n", code.mnemonic );
					break;
				}
			}
		}
		pos += code.size;
	}

	/* now we have the stack "just" revert */
	if ( st <= *rest_size ) {
		*rest_size = 0;
		while ( stack->size > 0 ) {
			for ( size = 0; size < stack->size; size++ ) {
				rest[(*rest_size)++] = stack->code[size];
			}
			tmp = stack;
			stack = tmp->prev;
			free( tmp );
		}
	} else {
		fatal( "Buffer to short (need %d)!\n", pos );
		err = -1;
	}
exit:
	disassemble_cleanup();
	return( err ); /* ? */
}
/* end of ripped code */
#endif

/* This is the new system using the libdisasm 0.21-pre1 */
/* Ripped from libdisasm x86dis.c :) */
#ifdef LIBDISASM_NO_COMPAT
void
kernsh_manual_print( x86_insn_t *insn, void *arg ) 
{
    char line[4096];
    int i;
	x86_op_t * op = NULL;
	
    if ( x86_format_insn(insn, line, 4096, (enum x86_asm_format)disasm_syn) <= 0 ) {
        return;
    }

    if ( disasm_syn == att_syntax || disasm_syn == intel_syntax  ) {
        /* print an address and hex bytes, since libdisasm doesn't */
       fprintf(stdout, "%#08lx ", insn->addr );
        for ( i = 0; i < 10; i++ ) {
            if ( i < insn->size ) {
                fprintf(stdout, "%02X", insn->bytes[i]);
            } else {
                fprintf(stdout, "  ");
            }
        }
        fprintf(stdout, "\t");
    }
	
	if ( insn->type == insn_call 
			|| insn->type == insn_callcc
			|| insn->type == insn_jmp 
			|| insn->type == insn_jcc 
		) {
		op = x86_get_dest_operand( insn );
		if ( op != NULL && (disasm_syn == att_syntax 
					|| disasm_syn == intel_syntax) ) {
			char sym[1024];
			char * tmp;
			memset(sym,0,sizeof(sym));
			x86_format_operand(op, insn, sym, sizeof(sym), (enum x86_asm_format)disasm_syn);
//			fprintf(stdout,"sym:%#lX",get_jmp_off(op,insn->addr+insn->size));
			if ( sym[0] != 0 ) {
				tmp = get_symbol_by_addr(sym);
				if ( ! (tmp[0] == '0' && tmp[1] == 'x') ) {
					snprintf(&line[strlen(line)-10], sizeof(line) - 10,
						   "%s", tmp );
				}
				free(tmp);
				tmp=NULL;
			}
		}
	}
    fprintf(stdout, "%s", line);
	fprintf(stdout, "\n");
}


void
display_code_asm( size_t start_address, unsigned char * buf, int len )
{
	int i = 0;
	x86_insn_t insn;
	int j = 0;
	char c = 0;
	
	x86_init(opt_none,kernsh_reporter);
	
	while ( i < len ) {
		if ( x86_disasm(buf, len, start_address, i, &insn) ){
			kernsh_manual_print( &insn, NULL );
			i += insn.size;
		} else {
			/* Should never occured !!! */
			i++;
		}

        if ( !script_on_command_line ) {
            j++;
            if ( !((j+1)%23) ) {
                c = enter();
                if ( c == '\r' || c == '\n' ) {
                    continue;
                }
                if ( c == 'q' ) {
                    getchar();
                    return;
                }
            }
        }	
	}
	x86_cleanup();
}
#endif
/* I have just modified a little bit the dissassembler to make it more
 * effecient with the addresses - Kstat
 */
#ifndef LIBDISASM_NO_COMPAT
void
display_code_asm( size_t start_address, unsigned char * buf, int len )
{
	unsigned char symbol_name[ 2*sizeof(size_t)+3 ];
	struct instr curr_inst;
	size_t p_symbol;
	char * tmp = NULL;
	int i = 0;
	int size;
	int n;
	int j = 0;
	int c;

	disassemble_init( IGNORE_NULLS, ATT_SYNTAX );

	while ( i < len ) {
		memset( &curr_inst, 0, sizeof(struct code) );
		fprintf( stdout, "%#x\t", start_address+i );
		/* Perhaps need to remove len - i */
		size = disassemble_address( buf+i, len - i ,&curr_inst );
		if ( size ) {
			for ( n = 0; n < 12; n++ ) {
				if ( n < size ) {
					fprintf( stdout, "%02x ", buf[i+n] );
				} else {
					fputs( "   ", stdout );
				}
			}
			fprintf( stdout, "%s", curr_inst.mnemonic );
			if ( curr_inst.dest[0] != 0 ) {
				/* Crazy isn't it ? Yes I learnt a lot making my own 
				 * disassembler ;) - Kstat 
				 */
				if ( (curr_inst.mnemType & (INS_TYPE_MASK)) == (INS_CALL) || (curr_inst.mnemType & (INS_TYPE_MASK)) == (INS_BRANCH) || (curr_inst.mnemType & (INS_TYPE_MASK)) == (INS_BRANCHCC) ) {
					if ( (curr_inst.destType & (OP_TYPE_MASK)) == (OP_REL) 
						&& (curr_inst.destType & (OP_PERM_MASK)) == (OP_X) ) {
						p_symbol = ( start_address + i ) + get_jmp_off( curr_inst );
						snprintf( symbol_name, sizeof(symbol_name), "%#x", p_symbol );
						tmp = get_symbol_by_addr(symbol_name) ;
						fprintf( stdout, "\t%s (r)", tmp );
						/* free get_symbol_by_addr */
						free(tmp);
						tmp=NULL;
					} else {
						/* We have a direct call (a register ?) */
						if ( (curr_inst.destType & (OP_TYPE_MASK)) == (OP_REG) ) {
							fprintf( stdout, "\t%s (d reg)", curr_inst.dest );
						} else {
							fprintf( stdout, "\t%p (d)", 
								(void *)strtoul(curr_inst.dest,NULL,16) );
						}
					}
				} else {
					fprintf( stdout, "\t%s", curr_inst.dest );
				}
			}
			if ( curr_inst.src[0] != 0 ) fprintf( stdout, ",\t%s", curr_inst.src );
			if ( curr_inst.aux[0] != 0 ) fprintf( stdout, " %s", curr_inst.aux );
			fputs( "\n", stdout );
/*			hexdump((char *)&curr_inst,sizeof(curr_inst),0);*/
			i += size;
		} else {
			fatal( "invalid opcode %02x\n", buf[i] );
			i++;
		}

		if ( !script_on_command_line ) {
			j++;
			if ( !((j+1)%60) ) {
				c = enter();
				if ( c == '\r' || c == '\n' ) {
					continue;
				}
				if ( c == 'q' ) {
					getchar();
					return;
				}
			}
		}
	}
	disassemble_cleanup(); /* Doh !! */
	return;
}

/* adapted from quikdis.c in libdisasm */
int
disasm( size_t start_address, int len )
{
	Elf32_Ehdr * target_header;
	Elf32_Phdr * program_header;
	struct stat tmpstat;
	void * image_vmlinux;
	unsigned char * buf = NULL;
	int entry;
	int ftarget;
	int n;
	int x;
	int offset;

	/* read ELF header */
	ftarget = found_vmlinux();
	if ( ftarget < 0 ) {
		return( -2 );
	}

	n = fstat( ftarget, &tmpstat );
	if ( n < 0 ) {
		perror( "disasm:fstat()" );
		return( -1 );
	}

	image_vmlinux = mmap( 0, tmpstat.st_size, PROT_READ, MAP_SHARED, 
	ftarget, 0 );
	if ( (int)image_vmlinux < 1 ) {
		perror( "disasm:mmap()" );
		return( -1 );
	}
	close( ftarget );

	target_header = image_vmlinux;

	/* read program header table */
	for ( x = 0; x < target_header->e_phnum; x++ ) {
		program_header = image_vmlinux + target_header->e_phoff + 
		( x * target_header->e_phentsize );

		/* If entry point is in this section */
		if ( target_header->e_entry >= program_header->p_vaddr && 
		target_header->e_entry <= 
		(program_header->p_vaddr + program_header->p_filesz)) {

			/* resolve entry point RVA to a file offset */
			entry = target_header->e_entry - 
			( program_header->p_vaddr - program_header->p_offset );

			/* read entire program segment into buffer */
			offset = start_address - target_header->e_entry;
			buf = image_vmlinux + entry + offset;
	
			display_code_asm( target_header->e_entry+offset, buf, len );
		}
	}

	munmap( image_vmlinux, tmpstat.st_size );
	disassemble_cleanup();
	return( 0 );
}
#endif
