/*
 * The project kernsh is (C) 2002-2004. This program is free software;
 * you can redistribute it and/or modify it under the terms of the GNU
 * General Public License as published by the Free Software Foundation;
 * Version 2. This guarantees your right to use, modify, and
 * redistribute this software under certain conditions.
 *
 * Source is provided to this software because we believe users have a
 * right to know exactly what a program is going to do before they run
 * it.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details (
 * http://www.gnu.org/copyleft/gpl.html ).
 *
 */

#include <linux/kernel.h>
#include <asm/unistd.h>
#include <errno.h>
#include <sys/utsname.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <stdarg.h>
#include "include/kernsh.h"

int be_paranoid = 0 ; // we are not paranoid by default isn't it ? */

/* This is a funny exercice :) */
/* Well, let think after some reflexion ;).
 * Consider that as not so usefull. Only mid-crackers will be defeated by this.
 * Why ? well, I use setvbuf to avoid them filtering output of sys_write(). 
 * So if they avoid this, even if I use fflush, nothing could be done. 
 * The only good way to know if what you see is the real output or not, is 
 * to display character by charater, and *sleep(1)* between each (Ouuchh).
 * So I will certainly add in future some level in paranoia :)
 * - Kstat (Hum, me parano ?? not at all :)
 */
int
my_fprintf(FILE *stream, const char * fmt, ...) {
	va_list ap;
	int ret = 0;
	if ( be_paranoid == 0 ) {
		
		va_start(ap,fmt);
		ret = vfprintf(stream,fmt,ap);
		va_end(ap);
	} else {
		// Should be sufficient
		char buf[16384];
		int i;
		va_start(ap,fmt);
		ret = vsnprintf(buf,sizeof(buf),fmt,ap);
		va_end(ap);
		if ( ret >= sizeof(buf)-1 )
		{
			fputs("Warning !!! Out of buffer\n",stdout);
		}
		for ( i = 0; i < ret;)
		{
			i += fwrite(buf+i,sizeof(char),1,stream); /* big question is : */
			/* Does this work ? (no automatic buffering ?) */
			fflush(stream);
		}
	}
	return ret;
}

/* need by sysinfo() */
_syscall1(int, sysinfo, struct sysinfo *, info);

int
get_number_cpus( void )
{
	size_t cpus;

	cpus = get_symbol_by_name( "smp_num_cpus" );
	if ( !cpus ) {
		return( 1 );
	}

	/* I don't have a multi-processor system to test smp_num_cpus */
	return( 666 );
}

void
usage_display_system_info( void )
{
    usage_t p_usage;

    p_usage.command_line =
        "display <no arguments>";
    p_usage.description =
		"This command makes it possible to display informations about your system\n"
		"as for example uptime or cpus..";
    p_usage.arguments = NULL;

    usage( &p_usage );
    return;
}

#define KVBASE_MASK ( 0x1fffff )
void
do_display_system_info( int argc, char * argv[] )
{
	char * cmdline = NULL;
	char cpuid[ BUFSIZ ];
	struct utsname p_uname;
	struct sysinfo s_info;
	size_t ldt = 0;
	size_t eax, ebx, ecx, edx;
	int size;
	int proc;
	
	uname( &p_uname );
    fprintf( stdout, "system name         : %s\n", p_uname.sysname );
    fprintf( stdout, "node name           : %s\n", p_uname.nodename );
    fprintf( stdout, "realease            : %s\n", p_uname.release );
    fprintf( stdout, "version             : %s\n", p_uname.version );
    fprintf( stdout, "machine name        : %s\n", p_uname.machine );
    fprintf( stdout, "domain name         : %s\n",
    p_uname.__domainname );
	fprintf( stdout, "using mmap r/w      : %s\n", mmap_use ? "yes" : "no" );
	fprintf( stdout, "using lkm r/w       : %s\n\n", lkm_use  ? "yes" : "no" );

    sysinfo( &s_info );
    fprintf( stdout, "uptime              : %lds\n", s_info.uptime );
    fprintf( stdout, "upload              : 1 min %ld / 5 min %ld / "
    "15 min %ld\n", s_info.loads[0], s_info.loads[1], s_info.loads[2] );
    fprintf( stdout, "RAM                 : total %ld / free %ld / "
    "shared %ld\n", s_info.totalram, s_info.freeram, s_info.sharedram );
    fprintf( stdout, "memory in buffers   : %ld\n", s_info.bufferram );
    fprintf( stdout, "swap                : total %ld / free %ld\n",
    s_info.totalswap, s_info.freeswap );
    fprintf( stdout, "process number      : %d\n", s_info.procs );
    fprintf( stdout, "cpus                : %d\n", get_number_cpus() );

	__asm__ volatile( "\tmovl $0, %%eax\n\tcpuid\n"
					: "=a" (eax), "=b" (ebx), "=c" (ecx), "=d" (edx)
					:
				);
	memcpy( &cpuid[0], &ebx, 4 );
	memcpy( &cpuid[4], &edx, 4 );
	memcpy( &cpuid[8], &ecx, 4 );
	cpuid[ 12 ] = 0;
	fprintf( stdout, "cpu ID string       : %s\n", cpuid );

    fprintf( stdout, "page size           : %d\n", getpagesize() );
    fprintf( stdout, "kernel virtual base : %#x\n",
    get_symbol_by_name("_stext") & ~KVBASE_MASK );

    fprintf( stdout, "system_call         : %#x\n", p_system_call );
    fprintf( stdout, "sys_call_table      : %#x\n", p_sys_call_table );
    fprintf( stdout, "idt                 : %#x\n", idtr.base );
    fprintf( stdout, "idt size            : %#x\n", idtr.limit );

	__asm__ volatile( "sgdt %0" : "=m" (gdtr) : );
	fprintf( stdout, "gdt                 : %#x\n", gdtr.base );
	fprintf( stdout, "gdt size            : %#x\n", gdtr.limit );

	__asm__ volatile( "sldt %0" : "=m" (ldt) : );
	fprintf( stdout, "ldt                 : %#x\n", ldt );

	/* get boot arguments */
	cmdline = malloc( BUFSIZ );
	if ( cmdline == NULL ) {
		perror( "do_display_system_info:malloc()" );
		return;
	}

	/* we suppose that sizeof(boot arguments) < BUFSIZ */
	memset( cmdline, '\0', BUFSIZ );
#define CMDLINE "/proc/cmdline"
	fprintf( stdout, "boot arguments      : " );
	proc = open( CMDLINE, O_RDONLY );
	if ( proc < 0 ) {
		fatal( "can't open %s file..\n", CMDLINE );
		goto out;
	}

	size = read( proc, cmdline, BUFSIZ );
	if ( size < 0 ) {
		perror( "do_display_system_info:read()" );
		goto out;
	}
	cmdline[ size-1 ] = '\0';

	fprintf( stdout, "%s\n", cmdline );
	free( cmdline );
out:
	close( proc );
	return;
}	

