/*
 * The project kernsh is (C) 2002-2004. This program is free software;
 * you can redistribute it and/or modify it under the terms of the GNU
 * General Public License as published by the Free Software Foundation;
 * Version 2. This guarantees your right to use, modify, and
 * redistribute this software under certain conditions.
 *
 * Source is provided to this software because we believe users have a
 * right to know exactly what a program is going to do before they run
 * it.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details (
 * http://www.gnu.org/copyleft/gpl.html ).
 *
 */

#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include "include/kernsh.h"

void
list_gdt( void )
{
		/*
	char * quad_word[] = 
		{ "NULL descriptor", 
		  "not used", 
		  "0x10 kernel 4GB code at 0x00000000",
		  "0x18 kernel 4GB data at 0x00000000",
		  "0x23 user   4GB code at 0x00000000",
		  "not used",
		  "not used",
		  "0x40 APM set up for bad BIOS's",
		  "0x48 APM CS code",
		  "0x50 APM CS 16 code (16 bit)",
		  "0x58 APM DS data",
		  "space for TSS's and LDT's"
		};
		*/
	unsigned char * buffer;
	unsigned int dp = 0;
	int i;
	int n;

	fprintf( stdout, "gdt address: %#x\n", get_symbol_by_name("gdt") );
	__asm__ volatile( "sgdt %0" : "=m" (gdtr) : );

	fprintf( stdout, "gdt table address: %#x\n", gdtr.base  );
	fprintf( stdout, "gdt size: %#x\n\n", gdtr.limit );

	buffer = malloc( 8 );

	fputs( "gdt table description :\n", stdout );
	for ( n = 0; n < gdtr.limit; n = n+8 ) {
		memset( buffer, '\0', sizeof(buffer) );
		i = read_kmem( buffer, gdtr.base+n, 8, file_descriptor_kmem );
		if ( i < 0 ) {
			return;
		}
	
		fputs( "  .quad 0x", stdout );
		for ( dp = 1; dp <= 8; dp++ ) {
			fprintf( stdout, "%02x", buffer[dp-1] );
		}
		fputs( "\n", stdout );
	}
	free( buffer ); /* Doh ! */
	return;
}

void
usage_gdt( void )
{
    usage_t p_usage;

    p_usage.command_line =
		"gdt [ -l ] <arguments>";
    p_usage.description =
		"This command makes it possible to handle the global description\ntable (or GDT).";
    p_usage.arguments =
		"  -l : display GDT.";

    usage( &p_usage );
    return;
}

void
do_gdt( int pargc, char * pargv[] )
{
	char c;
	
	NOARGUMENTS( pargc, 1, gdt );
	
	optind = 0;
	
	while ( (c = getopt(pargc, pargv, "l")) != EOF ) {
		switch ( c ) {
			case 'l': 
				list_gdt();
				break;
			default:
				usage_gdt();
				return;
		}
	}
	return;
}
