/*
 * The project kernsh is (C) 2002-2004. This program is free software;
 * you can redistribute it and/or modify it under the terms of the GNU
 * General Public License as published by the Free Software Foundation;
 * Version 2. This guarantees your right to use, modify, and
 * redistribute this software under certain conditions.
 *
 * Source is provided to this software because we believe users have a
 * right to know exactly what a program is going to do before they run
 * it.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details (
 * http://www.gnu.org/copyleft/gpl.html ).
 *
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/mman.h>
#include <string.h>
#include "include/kernsh.h"
#include "include/module.h"
#include <nrsys.h> /* Automatically created */

#define KMEM_FLAGS ( 0x20 + 0x10 + 0x40 + 0x80 + 0x100 )
#define KMALLOC_OBJ "=modules/lkm-kmalloc.o"
#define KFREE_OBJ "=modules/lkm-kfree.o"
#define PLACE_OBJ "=modules/lkm-place_rw.o"
#define STACK PAGE_OFFSET

#define SYSCALL_KMALLOC	NR_sysfree5
#define SYSCALL_KFREE	NR_sysfree6
#define SYSCALL_WLKM	NR_sysfree7
#define SYSCALL_RLKM	NR_sysfree8

/* These come from lkm-place_rw.c */
#define K_READ 0
#define K_WRITE 1
#define K_HIJACK 2

static int dry_run = 0;	/* If you don't want the read to be made ! */

struct kma_struct {
	unsigned long ( *kexec )( size_t, int );
	int size;
	int flags;
	unsigned long mem;
};

struct kfr_struct {
	unsigned long ( *kexec )( const void * );
	const void * address;
};

int
open_kmem( char * path_file_kmem )
{
	return( open(path_file_kmem, O_RDWR, 0) );
}

void
close_kmem( int kd )
{
	close( kd );
}

void *
my_memmem( void *heystack, size_t s_hey, void *needle, size_t s_nee, int joker ){
	if ( joker == NO_JOKER ) {
		return memmem(heystack,s_hey,needle,s_nee);
	} else {
		/* We have joker values ? \x00 now */
		return memmem_j(heystack,s_hey,needle,s_nee,(char)joker);
	}
}

size_t
kfind_area(size_t start, size_t end, void * stack, int size, int joker)
{
	return find_area(start,end,stack,size,joker,file_descriptor_kmem);
}

size_t
find_area( size_t start, size_t end, void * stack, int size, int joker, int fd ) 
{
	char * buf = NULL,*mm = NULL ;
	size_t addr=start;
	int n = 0;
	int len=0;
	if( (buf=malloc(BUFSIZ)) == NULL ) {
		perror( "find_area : malloc");
		return 0;
	}
	bzero(buf,BUFSIZ);
	for(;addr < end; addr+=BUFSIZ - size) {
		n = read_kmem(buf,addr,BUFSIZ,fd);
		if ( n == 0 ) {
			mm = my_memmem(buf,BUFSIZ,stack,(size_t)size,joker);
			if ( mm == NULL )
			{
				continue;
			}
			len = mm - buf;
			if (addr + len > end) {
				free(buf);
				return 0;
			}
			if (mm != NULL) {
				free(buf);
				return (size_t)(addr + len);
			}
		}
	}
	free(buf);
	return 0;
}

/* return 1 on success 0 else */
/* This function should perhaps be moved to module.c ?? */
int
launch_kamikaze( int what, size_t to, size_t addr )
{
	char cmd[128];
	char c = 0;
	switch ( what ) {
		case 0:
		case 1:
			sprintf(cmd,"insmod "PLACE_OBJ" where=%p what=%d",(void*)to,
				what);
			break;
		case 2:
			sprintf(cmd,"insmod "PLACE_OBJ" where=%p what=%d addr=%p",
					(void*)to, what, (void *)addr);
			fatal("this will use insmod/rmmod technique, it's possible that"
					"You didn't want this : type <q> to don't try this: ");
			c = enter();
			if ( c == 'q' ) {
				return 0;
			}
			break;
	}
	if ( system(cmd) == 0 )
	{
		/* XXX need to be able to change this :) */
		system("rmmod lkm-place_rw");
		return 1;
	}
	return 0;
}

/* Return 1 on success 0 else */
int
wfunc( size_t to, void* buf, size_t buf_size, int what )
{
	int n = 0;
	n = write_kmem(buf, to, buf_size, file_descriptor_kmem);
	if ( n == 0 ) {
		return 1;
	}
	if ( mmap_use == 0 )
	{
		mmap_use = 1;
		n = write_kmem(buf, to, buf_size, file_descriptor_kmem);
		mmap_use = 0;
		if ( n == 0 ) {
			return 1;
		}
	}
	return launch_kamikaze(what,to,0);
}

size_t old_stub_read = 0L;
size_t old_stub_write = 0L;

int
wfunc_write( size_t to /*, int chunk*/)
{
	char write_func[] = 
		"\x8b\x54\x24\x08"	/* mov    0x8(%esp,1),%edx */
		"\x8b\x44\x24\x04"	/* mov    0x4(%esp,1),%eax */
		"\x88\x10"			/* mov    %dl,(%eax) */
		"\xc3";				/* ret */
		if ( 0 == (old_stub_write = kfind_area(STACK + 0x00100000,
									get_symbol_by_name("high_memory"),
									write_func,sizeof(write_func) -1,NO_JOKER))) {
			old_stub_write = to;
			if ( lkm_use == 0 ) {
				return wfunc(to, write_func, sizeof(write_func) - 1, K_WRITE);
			} else {
				int n = 0;
				lkm_use = 0;
				n =  wfunc(to, write_func, sizeof(write_func) - 1, K_WRITE);
				lkm_use = 1;
				return n;
			}
		}
		return 1;
}

int 
wfunc_read( size_t to )
{
	char read_func[] = 
		"\x8b\x44\x24\x04"		/* mov    0x4(%esp,1),%eax */
		"\x0f\xb6\x00"			/* movzbl (%eax),%eax */
		"\xc3" ;				/* ret */
	if ( 0 == (old_stub_read = kfind_area(STACK + 0x00100000, 
									get_symbol_by_name("high_memory"), 
									read_func, sizeof(read_func) - 1,NO_JOKER)) ) {
		int n = 0;
		old_stub_read = to;
		if ( lkm_use == 1 ) {
			lkm_use = 0;
			n = wfunc(to, read_func, sizeof(read_func) - 1, K_READ);
			lkm_use = 1;
			return n;
		} else {
			return wfunc(to, read_func, sizeof(read_func) - 1, K_READ);
		}
	}
	return 1;
}

/* This function is responsible of creating the good way to create the 
 * read/write function of the kernel Kstat
 * We need to trust the first system ... If it lies byebye kernel :)
 */
size_t khijack_syscall(int,size_t,int);

int
create_rw_km( int what )
{
	size_t pos_to_place = 0;
	char * buf = malloc(BUFSIZ);
	if ( buf == NULL ) {
		perror("create_rw_km: malloc");
		return -1;
	}
	switch ( what ) {
		case 0:
			sprintf(buf,
					"\x90\x90\x90\x90\x90\x90\x90\x90"
				   );
			if ( 0 != (pos_to_place =
						kfind_area(STACK+0x00100000, get_symbol_by_name("high_memory"), buf,
							strlen(buf),NO_JOKER))) {
		/* We have the area needed,to we can now begin to work ;) */
		/* we have the choice, we can write it via the mmap/write function */
		/* or use a kamikaz module (I like this idea)*/
		/* we have a lot's of possibility in fact :) */
/*				fprintf(stdout,"put the read function to %p\n",
								(void*)pos_to_place);*/
				if (wfunc_read(pos_to_place) == 1) {
/*					fprintf(stdout, "Putted in fact here : %p\n",
									(void *)old_stub_read);*/
					return old_stub_read; 
				}
			} else {
				fprintf(stdout, "READ : Not found !\n");
				return 0;
			}
			break;
		case 1:
			/* Now the write function */
			sprintf(buf,		            
					"\x90\x90\x90\x90\x90\x90\x90\x90"
					"\x90\x90\x90"
				   );
			if (0 != (pos_to_place =
					kfind_area(STACK+0x00100000, get_symbol_by_name("high_memory"),
							buf, strlen(buf),NO_JOKER))) {
/*				fprintf(stdout,"put the write function to %p\n",
								(void*)pos_to_place);*/
				if (wfunc_write(pos_to_place) == 1) {
/*						fprintf(stdout,"Putted in fact here : %p\n",
										(void *)old_stub_write);*/
					return old_stub_write;
				}
			} else {
				fprintf(stdout, "WRITE : Not found !\n");
				old_stub_write = 0L;
				return 0;
			}
			break;
		default:
			break;
	}
	return 0;	
}

int
read_from_lkm(void * buf,size_t addr, int buf_size, int fp)
{
	int __ret=0;
	size_t old_addr;
	unsigned int argv[1];
	if ( old_stub_read == -1 || old_stub_read == 0)	{
		return -1;
	}
	if ( buf_size <= 0 ) {
		return 0;
	}
	if( -1 == (old_addr = khijack_syscall(SYSCALL_RLKM, old_stub_read,
					file_descriptor_kmem))) {
		return -1;
	}
	while( buf_size-- != 0 ) {
		argv[0] = addr++;
		*(char*)(buf++) = (char)scall(SYSCALL_RLKM,1,argv);
		__ret++;
	}
	khijack_syscall(SYSCALL_RLKM, old_addr, file_descriptor_kmem);
	return 0;
}

size_t old_addr_write= 0L;
int hijack_done = 0;

int
write_from_lkm(void * buf,size_t addr, int buf_size, int fp)
{
	int __ret=0;
	unsigned int argv[2];
	if ( old_stub_write == -1 || old_stub_write == 0) {
		return -1;
	}
	if ( buf_size <= 0 ) {
		return -1;
	}
	/* We have a design decision to take here ...
	 * First we can set the syscall one time and let it like this in
	 * memory (logical)
	 * Second we can set up each time the syscall but we need to write 
	 * with another technik (mmap/write/...) So ould be a problem depends 
	 * on the needs ...
	 * Ahem... there is a loop here ;), so need to  disable the write :/
	 */
	lkm_use = 0;
	if ( hijack_done == 0 && old_addr_write == 0 && -1 == (old_addr_write = 
							khijack_syscall(SYSCALL_WLKM,  old_stub_write,
											file_descriptor_kmem))) {
		lkm_use = 1;
		hijack_done = 0;
		return -1;
	} else {
		hijack_done = 1;
	}
	lkm_use = 1;
	do {
		argv[0] = addr++;
		argv[1] = *(unsigned char*)(buf++);
		scall(SYSCALL_WLKM,2,argv);
/*		fprintf(stdout,"get %p %02x\n",(void *)argv[0],(unsigned char)argv[1]);*/
		__ret++;
	} while ( --buf_size != 0 );
	/* Ok, writing if a definitive solution 'till exit or kmem -z  */
/*	lkm_use = 0;
	khijack_syscall(SYSCALL_WLKM,  old_addr_write, file_descriptor_kmem);
	lkm_use = 1;*/
/*	old_addr = 0L;*/
	return __ret; 
}

size_t
round_the_offset( size_t offset )
{
	size_t oft = 0;

	oft = ( offset - (size_t)PAGE_OFFSET ) -
		( (offset-(size_t)PAGE_OFFSET) & ~4095 ); /* better ? & ~(getpagesize() - 1) */
	return( oft );
}

/* An alternative to writing/reading kmem. Original Idea by Guillaume
 * Pelat <pelat_g _ at _ epita.fr> modifying for match needs of kernsh. See
 * kmem_mmap.gz... 
 */
int
mmap_write_kmem( void * buf, size_t offset, int size, int kd )
{
	size_t oft = 0;
	char * mem = NULL;
	int i;
	
	/* first we need to round the offset to the x of 4096 due to the round in mmap */
	oft = round_the_offset( offset );

	mem = mmap( 0, oft+(size_t)size, PROT_WRITE, MAP_SHARED, kd, 
		((offset-(size_t)PAGE_OFFSET) & ~4095) );
	if ( mem == MAP_FAILED ) {
		fatal( "can't map memory in write mode !\n" );
		return( -1 );
	}

	memcpy( mem+oft, buf, size );

	/* this is not really needed because mumap do the same */
	i = msync( mem, oft+size, MS_SYNC | MS_INVALIDATE );
	if ( i == -1 ) {
		fatal( "can't write in the kernel !\n" );
		return( -1 );
	}
	i = munmap ( mem, oft+size );
	if ( i == -1 ) {
		fatal( "can't unmap the memory !\n" );
		return( -1 );
	}
	return( 0 );
}

int
mmap_read_kmem( void * buf, size_t offset, int size, int kd )
{
	size_t oft = 0;
	char * mem = NULL;
	int i;

	/* first we need to round due to the round in mmap*/
	oft = round_the_offset( offset );

	mem = mmap( 0, oft+(size_t)size, PROT_READ, MAP_SHARED, kd,
		((offset-(size_t)PAGE_OFFSET) & ~4095) );
	if ( mem == MAP_FAILED ) {
		fatal( "can't map memory in read mode !\n" );
		return( -1 );
	}

	memcpy( buf, mem+oft, size );
	
	i = munmap( mem, oft+size );
	if ( i == -1 ) {
		fatal( "can't unmap the memory !\n" );
		return( -1 );
	}
	return( 0 );
}

int
write_kmem( void * buf, size_t offset, int size, int kd )
{
	int n;
	if ( dry_run == 1 )
	{
		fprintf(stdout,"Writing :\n");
		hexdump(buf,size,offset);
		return 0;
	} 
	if ( mmap_use == 1 && kd == file_descriptor_kmem ) {
		return( mmap_write_kmem(buf, offset, size, kd) );
	}
	if ( lkm_use == 1 && kd == file_descriptor_kmem && old_stub_write != 0) {
		n = write_from_lkm(buf, offset, size, kd);
	} else {
		n = lseek( kd, offset, SEEK_SET );
		if ( n != offset ) {
			perror( "write_kmem:lseek()" );
			return( -1 );
		}

		n = write( kd, buf, size );
	}
	if ( n != size ) {
		perror( "write_kmem:write()" );
		return( -1 );
	}
	return( 0 );
}

int
read_kmem( void * buf, size_t offset, int size, int kd )
{
	int n;

	if ( mmap_use == 1 && kd == file_descriptor_kmem ) {
		return( mmap_read_kmem(buf, offset, size, kd) );
	}
	
	if ( lkm_use == 1 && kd == file_descriptor_kmem && old_stub_read != 0) {
		return( read_from_lkm(buf, offset, size, kd) );
	}

	n = lseek( kd, offset, SEEK_SET );
	if ( n != offset ) {
		perror( "read_kmem:lseek()" );
		return( -1 );
	}
	
	n = read( kd, buf, size );
	if ( n < 0 ) {
		perror( "read_kmem:read()" );
		return( -1 );
	}
	return( 0 );
}

void
kdump( size_t addr, int ssize, int file_descriptor )
{
	char * c = NULL;
	int n;

	c = ( char * )malloc( ssize );
	if ( c == NULL ) {
		perror( "kdump:malloc()" );
		return;
	}

	memset( c, '\0', ssize );
	n = read_kmem( c, addr, ssize, file_descriptor );
	if ( n < 0 ) {
		return;
	}

	hexdump( c, ssize, addr );
	free( c );

	return;
}

/* why disassemble the bzimage, when you can do it directly from hell ?
 * So you can track directly this f*** kernel horses ;)
 */
void
kdisass( size_t addr_s, int size, int file_descriptor )
{
	unsigned char * buf = NULL;
	int n = 0;

	buf = (unsigned char *)calloc( size, sizeof(char) );
	if ( buf == NULL ) {
		perror( "kdisass:calloc()" );
		return;
	}

	n = read_kmem( buf, addr_s, size, file_descriptor );
	if ( n < 0 ) {
		return;
	}

	display_code_asm( addr_s , buf, size );
	free( buf );
	return;
}

size_t
khijack_syscall( int sysnr, size_t new_address, int fd )
{
	size_t save_address;
	int i = sysnr;
	int n;

	/*for ( save_address = i = 0; i != NR_syscalls; i++ ) {
        if ( i == sysnr ) {*/
			save_address = sys_call_table[i];
			sys_call_table[i] = new_address;
/*			break;
		}
	}*/
	n = write_kmem( &sys_call_table[i], p_sys_call_table+i*sizeof(sys_call_table[0]),
	sizeof(sys_call_table[0]), fd );
	if ( n < 0 ) {
		/* Trying to use the lkm-place_rw technique */
		if ( launch_kamikaze(K_HIJACK, p_sys_call_table + i *
					sizeof(sys_call_table[0]), sys_call_table[i] ) == 0 ) {
			fatal( "can't khijack() syscall, aborting !\n" );
			return( -1 );
		}
	}
	return( save_address );
}

int
build_structure_kkmalloc( int sz, int nbr_syscall, 
unsigned long * address )
{
	struct kma_struct kmalloc;
	unsigned int arg[1];

	kmalloc.kexec = ( void * )get_symbol_by_name( "kmalloc" );
	if ( !kmalloc.kexec ) {
		fatal( "can't find symbol kmalloc..\n\n" );
		return( -1 );
	}

	kmalloc.size = sz;
	kmalloc.flags = KMEM_FLAGS;
	kmalloc.mem = 0;

	arg[0] = (unsigned int)&kmalloc;
	scall( nbr_syscall, 1, arg );
	if ( !kmalloc.mem ) {
		fatal( "kmalloc failed..\n" );
		return( -1 );
	}
	
	sprintf( env.value, "%#08x", (unsigned int)kmalloc.mem );
   	fprintf( stdout, "memory malloc()ed at %#x\n", 
		(unsigned int)kmalloc.mem );
	*address = kmalloc.mem;
	return( 0 );
}

void
kkmalloc( int ssize, unsigned long * address )
{
    char save[ BUFSIZ ];
    size_t addr_save;
	size_t m;
    int n;
    
    n = read_kmem( &save, STACK, 192, file_descriptor_kmem );
	if ( n < 0 ) return;
    n = kinsert( KMALLOC_OBJ, STACK, 0, 0 );
    if ( n < 0 ) return;

	addr_save = khijack_syscall( SYSCALL_KMALLOC, STACK, 
		file_descriptor_kmem );
	if ( addr_save > 0 ) {
		n = build_structure_kkmalloc( ssize, SYSCALL_KMALLOC, address );
		if ( n < 0 ) return;
		m = khijack_syscall( SYSCALL_KMALLOC, addr_save, 
			file_descriptor_kmem );
		if ( m < 0 ) return;
		write_kmem( &save, STACK, 192, file_descriptor_kmem );
	}
    return;
}

int
build_structure_kkfree( size_t address, int nbr_syscall )
{
	struct kfr_struct kfree;
	unsigned int arg[1];
	
	kfree.kexec = ( void * )get_symbol_by_name( "kfree" );
	if ( !kfree.kexec ) {
		fatal( "can't find symbol kfree..\n\n" );
		return( -1 );
	}

	kfree.address = (const void *)address;
	
	arg[0] = (unsigned int)&kfree;
	scall( nbr_syscall, 1, arg );
	sprintf( env.value, "0" );
	fprintf( stdout, "memory to %#x free()ed\n\n", address );
	return( 0 );
}

void
kkfree( size_t address )
{
    char save[ BUFSIZ ];
    size_t address_save;
    size_t m;
    int n;

    n = read_kmem( &save, STACK, 192, file_descriptor_kmem );
	if ( n < 0 ) return;
    n = kinsert( KFREE_OBJ, STACK, 0, 0 );
	if ( n < 0 ) return;
	address_save = khijack_syscall( SYSCALL_KFREE, STACK, 
		file_descriptor_kmem );
	if ( address_save < 0 ) return;
	n = build_structure_kkfree( address, SYSCALL_KFREE );
	if ( n < 0 ) return;
	m = khijack_syscall( SYSCALL_KFREE, address_save, 
		file_descriptor_kmem );
	if ( m < 0 ) return;
	write_kmem( &save, STACK, 192, file_descriptor_kmem );

	return;
}

void 
kinsmod( char * module, char * function )
{
	unsigned long address;
	size_t func_addr;
	size_t address_save;
	size_t m;
	int __ret;
	int len ;
	int lkm_file_desc;
	int n;

	lkm_file_desc = open( module, O_RDONLY );
	if ( lkm_file_desc < 0 ) {
		fatal( "%s : No such module..\n\n" );
		return;
	}

	len = module_size( lkm_file_desc );
	close( lkm_file_desc );
	if ( len < 0 ) {
		return;
	}

    kkmalloc( len, &address );
    if ( address <= 0 ) {
		fatal( "can't kkmalloc() memory.. aborting !\n" );
		return;
	}
	
    n = kinsert( module, (size_t)address, function, &func_addr );
    if ( n < 0 ) {
		return;
	}

    address_save = khijack_syscall( SYSCALL_KFREE, func_addr, 
	file_descriptor_kmem );
	if ( address_save < 0 ) {
		return;
	}

    SYSCALL0( SYSCALL_KFREE );

	m = khijack_syscall( SYSCALL_KFREE, address_save, 
	file_descriptor_kmem );
	if ( m < 0 ) {
		return;
	}
    fprintf( stdout, "insertion at\n%#x and %s() executed !\n\n", 
		module, (int)address, function );
	return;
}

void
krmmod( char * m, size_t addr, char * function )
{
	size_t func_addr = 0;
	size_t address_save, n;
	int __ret;
	int len;
	void * module = NULL ;

	module = rel_module( m, addr, &len, function, (size_t *)&func_addr );
	free( module );
	if ( func_addr < STACK ) {
		fatal( "Function not found !\n" );
		return;
	}

	address_save = khijack_syscall( SYSCALL_KFREE, func_addr, 
		file_descriptor_kmem );
	if ( address_save < 0 ) {
		return;
	}

	SYSCALL0( SYSCALL_KFREE );
	n = khijack_syscall( SYSCALL_KFREE, address_save, file_descriptor_kmem );
	if ( n < 0 ) {
		return;
	}

	/* So the module could be freed now */
	kkfree( addr );
	return;
}

int
kinsert( char * m, size_t location, char * function, size_t * addr )
{
	void * module;
	int len;
	int n;

	module = rel_module( m, location, &len, function, addr );
	if ( !module ) { 
		fatal( "cannot make relocation for %s..\n", m );
		return( -1 );
	}

	n = write_kmem( module, location, len, file_descriptor_kmem );
	if ( n < 0 ) {
		fatal( "error, module %s NOT inserted into /dev/kmem..\n", m );
		return( -1 );
	}	

	fprintf( stdout, "module %s inserted in /dev/kmem..\n", m );
	free(module);
	return( 0 );
}

void
kscan( size_t lower, size_t upper, char * string, size_t size, int joker ,int fd )
{
	char *mm = NULL;
	int j = 0;
	int p = 0;
	size_t prev = 0x0;

	fputs( "scanning all kernel data... wait :\n\n", stdout );
	mm = (char *)lower;
	do {
		prev = (size_t)( mm );
		mm = (void *)find_area(prev, upper, string, size, joker, fd);
		if ( mm != NULL ) {
			fprintf( stdout, "pattern found at %p ", mm );
			fprintf( stdout, "\n" ); /* or mm - prev :) */
			if ( !script_on_command_line ) {
				j++;
				if ( j == 23 ) {
					char c = enter();
					if ( c == 'q' ) {
						return;
                    }
					j = 0;
				}
            }
			fflush( stdout );
			p++;
			mm += size;
		}
	} while ( mm != NULL );
	
	if ( p == 0 ) {
		fatal( "no pattern found\n" );
	} else {
		fprintf( stdout, "found %d pattern \n", p );
	}
	return;
}

void
kcopy( size_t source, size_t destination, int ssize )
{
	char * cache = NULL;
	int n;
	
	cache = malloc( ssize + 2 ); /*XXX +2 because we want ssize+1 !? */
	if ( cache == NULL ) {
		perror( "kcopy:malloc()" );
		return;
	}
	
	n = read_kmem( cache, source, ssize, file_descriptor_kmem );
	if ( n == 0 ) {
		cache[ssize+1] = '\0'; /* why +1 ??? and why '\0' ? */
		n = write_kmem( cache, destination, ssize, file_descriptor_kmem );
		if ( n < 0 ) {
			return;
		}
	}
	free( cache );
	return;
}

void
kwrite( size_t location, char * string )
{
	char * str2hex = NULL;
	char c;
	int i;
	int n;
	
	str2hex = string2hex( string, &i );
	if ( str2hex == NULL ) {
		return;
	}

	fputs( "\n[ before ]\n", stdout );
	kdump( location, i, file_descriptor_kmem );
	fputs( "\n[ after ]\n", stdout );
	hexdump( str2hex, i, location );
	fputs( "\nreally write [y/n] ? ", stdout );
	c = getchar();
	getchar();
	if ( c == 'y' || c == 'Y' ) {
		n = write_kmem( str2hex, location, i, file_descriptor_kmem );
		if ( n == 0 ) {
			fprintf( stdout, "string %s written in /dev/kmem\n", 
				string );
		}
	} else {
		fatal( "don't write in /dev/kmem..\n" );
	}
	free( str2hex );
	return;
}

void
toggle_lkm_use ( void )
{
	/* This is lazy ... 'cose if one of this function fail, when
	 * switching off, the system attempt to rebuild this
	 * and we cannot check against it due to the -l flag ...
	 */
	if ( old_stub_read == 0 && old_stub_read != -1) {
		old_stub_read = create_rw_km( 0 );
	}
	if ( old_stub_write == 0 && old_stub_write != -1) {
		old_stub_write = create_rw_km( 1 );
	}
	if ( lkm_use > 0 ) {
		lkm_use --;
		khijack_syscall(SYSCALL_WLKM,  old_addr_write, file_descriptor_kmem);
		old_addr_write = hijack_done = 0;
	} else {
		lkm_use ++;
	}
	return;
}

void
usage_kmem( void )
{
    usage_t p_usage;

    p_usage.command_line =
		"kmem [ -mfdicwkulxgrosj ] <arguments>";
    p_usage.description =
		"This command makes it possible to handle the device /dev/kmem\n"
		"with its own way, to read there, to write there, and mainly to\n"
		"include there files objects.";
    p_usage.arguments =
		"  -m [size] : allocate memory capacity /dev/kmem ;\n"
		"  -f [address] : free an allocated memory capacity ;\n"
		"  -d [address] [size] [file] : display a block of file\n" 
		"<file> (/dev/kmem ou /dev/mem) according to the offset and the size ;\n"
		"  -g [address] [size] [file] : dissasemble this block ;\n"
		"  -i [module] [address] <symbol_to_get_addr>: insert a module in \n"
		"/dev/kmem at address <address> ;\n"
		"  -k [module] [function] : insmod <module> in /dev/kmem\n"
		"<function> will be executed (commonly init_module) ;\n"
		"  -r [module] [module_address] [function] : rmmod [module_address]\n"
		"and calling before [function] (commonly cleanup_module) ;\n"
		"  -s [start address] [end address] [string] [file] : search a string\n"
		"in <file> (/dev/kmem ou /dev/mem) ;\n"
    	"  -o [start address] [end address] [opcode] [file] : search opcode\n"
		"in <file> (/dev/kmem ou /dev/mem) ;\n"
		"  -j [character] : WITH -s or -o option, give a joker character of your\n"
		"your choice ( like ? or \\x00 ) ;\n"
		"  -u : toggle the use of mmap_kmem or write_kmem (for kernel only);\n"
		"  -l : toggle the use of lkm read/write (for extended kernel only);\n"
		"  -x : display what I want to write instead of doing it :) \n"
		"  -c [source address] [destination address] [size] : copy a block\n"
		"of /dev/kmem from an address towards another ;\n"
		"  -w [address] [string] : write a string in /dev/kmem\n"
		"(e.g. -w 0xc0000000 \\x41\\x42\\x43\\x44).";

    usage( &p_usage );
    return;
}	

enum {
	WRITE_IN_KMEM,
	DUMP_KMEM,
	INSERT_IN_KMEM,
	SCAN_KMEM,
	COPY_KMEM,
	INSMOD_IN_KMEM,
	RMMOD_IN_KMEM,
	DISS_KMEM,
	HOOK_KMEM
};

void
do_kmem( int pargc, char * pargv[] )
{
	char * file = NULL;
	char c;
	char * buf2hex = NULL; /* for string2hex */
	size_t l;
	size_t ll;
	int ssize;
	int flag;
	int file_descriptor_mem;
	unsigned long kma;
	int si = 0;
	int joker = NO_JOKER;
	
	NOARGUMENTS( pargc, 1, kmem );

	ssize = 0;
	flag = -1;
	file = NULL;
	optind = 0;

	while ( (c = getopt(pargc, pargv, "uidcskwgrom:j:f:lx")) != EOF ) {
		switch ( c ) {
		case 'x':
			if ( dry_run > 0 ) {
				dry_run--;
			} else {
				dry_run++;
			}
			break;
		case 'l':
			toggle_lkm_use();
			break;
		case 'u':
			if ( mmap_use > 0 ) {
				mmap_use--;
			} else {
				mmap_use++;
			}
			break;
		case 'm':
			ssize = atoi( optarg );
			/*
			 if ( ssize < 0 || ssize > 1024 ) {
			 	fatal( "size must be > 0 and < 1024..\n" );
			 	return;
			 }
			 */
			kkmalloc( ssize, (unsigned long *)&kma );
			break;
		case 'f':
			XSTRTOUL( optarg, l );
			kkfree( l );
			break;
        case 'w':
            flag = WRITE_IN_KMEM;
            break;
        case 'd':
            flag = DUMP_KMEM;
            break;
        case 'i':
            flag = INSERT_IN_KMEM;
            break;
        case 'k':
            flag = INSMOD_IN_KMEM;
            break;
		case 'r':
			flag = RMMOD_IN_KMEM;
			break;
		case 'j':
			joker = 0;
			if ( strlen(optarg) == 1 ) {
				(char)joker = optarg[0];
			} else if ( strlen(optarg) == 4 ) {
				char * buf = string2hex( optarg, &ssize );
				if ( buf != NULL ) {
					(char)joker = buf[0];
					fprintf(stdout,"%02x\n",(unsigned char)joker);
					free( buf );
					buf = NULL;
				} else {
					fatal("can't get the joker !\n");
					return;
				}
			} else {
				fatal("invalid joker string !\n");
				return ;
			}
			break;
        case 's':
			if ( pargc > 4 ) {
				si = strlen(pargv[optind+2]);
			}
            flag = SCAN_KMEM;
            break;
        case 'c':
            flag = COPY_KMEM;
            break;
		case 'o':
			if ( pargc > 4 ) {
				si = (int)strlen(pargv[optind+2])/4; /* sizeof("\x00") - 1 */
				buf2hex = string2hex( pargv[optind+2], &ssize );
				if ( buf2hex == NULL ) {
					return;
				}
			}
			flag = SCAN_KMEM;
			break;
		case 'g':
			flag = DISS_KMEM;
			break;
        default:
            usage_kmem();
            return;
        }
    }	

    switch ( flag ) {
    case WRITE_IN_KMEM:
        if ( pargc != 4 ) {
            usage_kmem();
            return;
        }

        XSTRTOUL( pargv[2], l );
        kwrite( l, pargv[3] );
        break;

    case DUMP_KMEM:
        if ( pargc < 4 ) {
            usage_kmem();
            return;
        }

        XSTRTOUL( pargv[2], l );
        ssize = atoi( pargv[3] );
			
		fprintf( stdout, "\n[ %#x -> %#x ]\n", l, l+ssize );
        switch ( pargc ) {
        case 4:
            kdump( l, ssize, file_descriptor_kmem );
            break;
        case 5:
            file_descriptor_mem = open( pargv[4], O_RDONLY );
            if ( file_descriptor_mem < 0 ) {
                fatal( "can't open %s\n", pargv[4] );
                return;
            }
            kdump( l, ssize, file_descriptor_mem );
			close( file_descriptor_mem );
        }
        break;

	case DISS_KMEM:
		if ( pargc < 4 ) {
			usage_kmem();
			return;
		}

		XSTRTOUL( pargv[2], l );
		ssize = atoi( pargv[3] );

		fprintf( stdout, "\n[ %#x -> %#x ]\n", l, l+ssize );
		switch ( pargc ) {
		case 4:
			kdisass( l, ssize, file_descriptor_kmem );
			break;
		case 5:
			file_descriptor_mem = open( pargv[4], O_RDONLY );
			if ( file_descriptor_mem < 0 ) {
				fatal( "Can't open %s\n", pargv[4] );
				return;
			}
			kdisass( l, ssize, file_descriptor_mem );
			close( file_descriptor_mem );
		}
		break;

    case INSERT_IN_KMEM:
        if ( pargc < 4 || pargc > 5 ) {
            usage_kmem();
            return;
        }

        XSTRTOUL( pargv[3], l );
		switch (pargc) {
			case 4:
		        kinsert( pargv[2], l, 0, 0 );
				break;
			case 5:
				kinsert( pargv[2], l, pargv[4], &l );
				fprintf( stdout, "%s located at %#08x\n", pargv[4], l );
				break;
			default:
				break;
		}
        break;

	case RMMOD_IN_KMEM:
		if ( pargc != 5 ) {
			usage_kmem();
			return;
		}
		XSTRTOUL( pargv[3], l );
		krmmod( pargv[2], l, pargv[4] );
		break;

    case INSMOD_IN_KMEM:
        if ( pargc != 4 ) {
            usage_kmem();
            return;
        }

        kinsmod( pargv[2], pargv[3] );
        break;
    case SCAN_KMEM:
        if ( pargc < 5 ) {
            usage_kmem();
            return;
        }

        XSTRTOUL( pargv[optind], l );
        XSTRTOUL( pargv[optind+1], ll );
		if ( ll - l <= 0 ) {
			fatal( "size of memory area you want to scan "
				"is negative or null !\n" );
			return;
		} 
        switch ( pargc ) {
		case 7:
        case 5:
			if ( buf2hex == NULL ) {
	           	kscan( l, ll, pargv[optind+2], si, joker, 
						file_descriptor_kmem );
			} else {
				kscan( l, ll, buf2hex, si, joker, file_descriptor_kmem );
				free(buf2hex);
				buf2hex = NULL;
			}	
            break;
		case 8:
        case 6:
            file_descriptor_mem = open( pargv[optind+3], O_RDONLY );
            if ( file_descriptor_mem < 0 ) {
                fatal( "can't open %s\n", pargv[optind+3] );
                return;
            }
			if ( buf2hex == NULL ) {
	            kscan( l, ll, pargv[optind+2], si, joker,
					   	file_descriptor_mem );
			} else {
				kscan( l, ll, buf2hex, si, joker, file_descriptor_mem );
				free(buf2hex);
				buf2hex = NULL;
			}	
			close(file_descriptor_mem); /* Yeah close it !!! */
        }
        break;

    case COPY_KMEM:
        if ( pargc != 5 ) {
            usage_kmem();
            return;
        }

        XSTRTOUL( pargv[2], l );
        XSTRTOUL( pargv[3], ll );
        kcopy( l, ll, atoi(pargv[4]) );
        break;

    default:
        return;
    }

    return;
}

