/*
 * The project kernsh is (C) 2002-2004. This program is free software;
 * you can redistribute it and/or modify it under the terms of the GNU
 * General Public License as published by the Free Software Foundation;
 * Version 2. This guarantees your right to use, modify, and
 * redistribute this software under certain conditions.
 *
 * Source is provided to this software because we believe users have a
 * right to know exactly what a program is going to do before they run
 * it.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details (
 * http://www.gnu.org/copyleft/gpl.html ).
 *
 */

#include <dirent.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include "include/kernsh.h"
#include "../libdisasm/libdis.h"

#define PROC "/proc"

/* Return 0 if not hidden
 * 1 if well hidden ( ala adore )
 * 2 if not well hidden ( ala suckkit )
*/
int
is_hided( pid_t pid )
{
	char name[ 42 ];
	DIR * dir;
	struct dirent **namelist;
	int n;

	sprintf( name, PROC"/%d/", pid );
	dir = opendir( name );
	if ( !dir ) {
		return( 1 );
	}

	closedir( dir );
	n = scandir( PROC, &namelist, 0, alphasort );
	if ( n < 0 ) {
		perror ( "is_hided:scandir()" );
	} else {
		while ( n-- ) {
			if ( pid == atoi(namelist[n]->d_name) ) {
				return( 0 );
			}
		}
		if ( n <= 0 ) {
			return( 2 );
		}
	}

	return( 0 );
}

void 
usage_process( void )
{
	usage_t p_usage;

	p_usage.command_line = 
		"process [ -ifspch ] <arguments>";
	p_usage.description =
		"This command manipulate process structure (struct task_struct)\n"
		"and their sub-structure.";
	p_usage.arguments =
		"  -i [pid] : display information about the processus [pid] ;\n"
		"  -s [pid] : display the task_structure of processus [pid] ;\n"
		"  -f [pid] : display file descriptor list of processus [pid] ;\n"
		"  -h [pid] : Hide processus [pid] ;\n"
		"  *** Warning ! kernel will crash when process will exit ! ;\n"
		"  -p : display the whole list of processus in the 'ps' style ;\n" 
		"  -c : check for hided processus (this can be defeat).";

    usage( &p_usage );
    return;
}

typedef enum { ERR_TYPE, INFO_TYPE, TS_TYPE, FILE_TYPE, PS_TYPE, HIDE_TYPE,
HIDE_DETECT_TYPE } param;

void 
do_process( int pargc, char * pargv[] )
{
	pid_t pid = 0;
	param type;
	int ch;

	NOARGUMENTS( pargc, 1, "process" );

	type = ERR_TYPE;
	optind = 0;

    while ( (ch = getopt(pargc, pargv, "pci:s:f:h:")) != EOF ) {
		switch ( ch ) {
			case 'i' :
			case 's' :
			case 'f' :
			case 'h' :
				type = ( ch == 'i' ? INFO_TYPE : ch == 's' ? TS_TYPE : 
						 ch == 'f' ? FILE_TYPE : HIDE_TYPE );
				pid = strtoul( optarg, 0, 0 );
				break;
			case 'p' :
				type = PS_TYPE;
				break;
			case 'c' :
				type = HIDE_DETECT_TYPE;
				break;
			default :
				usage_process ();
				return;
		}
	}

	switch ( type )  {
		case ERR_TYPE :
			fatal( "unknow internal error !\n" );
			break;
		case PS_TYPE :
			proc_kps();
			break;
		case HIDE_DETECT_TYPE :
			proc_khide_detect();
			break;
		case INFO_TYPE :
			proc_kinfo( pid );
			break;
		case TS_TYPE :
			proc_kts( pid );
			break;
		case FILE_TYPE :
			proc_kfile( pid );
			break;
        case HIDE_TYPE  :
			proc_khide( pid );
			break;
		default :
			return;
	}
	
	return;
}

