/*
 * The project kernsh is (C) 2002-2004. This program is free software;
 * you can redistribute it and/or modify it under the terms of the GNU
 * General Public License as published by the Free Software Foundation;
 * Version 2. This guarantees your right to use, modify, and
 * redistribute this software under certain conditions.
 *
 * Source is provided to this software because we believe users have a
 * right to know exactly what a program is going to do before they run
 * it.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details (
 * http://www.gnu.org/copyleft/gpl.html ).
 *
 */

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include "include/kernsh.h"
#include <sysc.h> /* in =build/`uname -r`*/

int
set_system_call( void )
{
	int n = 0;

	/* get idt */
#ifndef DEBUG
	asm( "sidt %0" : "=m" (idtr) );
	/* get system_call address */
	n = read_kmem( &idt, idtr.base + 8 * 0x80, sizeof(idt),
		file_descriptor_kmem );
	if ( n < 0 ) {
		return( n );
	}
	p_system_call = (idt.off2 << 16) | idt.off1;
#else
	n = 0;
	p_system_call = get_symbol_by_name( "system_call" );
#endif

	if ( p_system_call != get_symbol_by_name("system_call") ) {
		fprintf( stdout, "\n[#] The IDT is hijacked ! Fake system_call"
			" at %#x\n", p_system_call );
		fputs( "[#] Your System.map file is perhaps bad\n", stdout );
		fputs( "[#] Create it with the script scripts/create_system_map.sh\n", stdout );
		if ( force_flag ) {
			p_system_call = get_symbol_by_name( "system_call" );
			fprintf( stderr, "[#] Forcing the symbol at  %#x\n\n", p_system_call );
		} else {
			fprintf( stderr, "[#] Try to use the -f option if kernsh fail !\n\n" );
		}
		return( 1 );
	}
	return( 0 );
}

int
set_syscall_table( void )
{
	int n = 0;
	size_t offset = 0;
	char * here = NULL;
	char buf[ BUFSIZ ];

	/* get sys_call_table address */
	/* the following method should work for 2.2/2.4/2.6 kernel */
	n = read_kmem( buf, p_system_call, sizeof(buf), file_descriptor_kmem );
	if ( n < 0 ) {
		return( -1 );
	}

	here = (char *)memmem( buf, sizeof(buf), "\xff\x14\x85", 3 );
	if ( here == NULL ) {
		perror( "set_syscall_table:memmem()");
		return( -1 );
	}
	offset = (int)( here - p_system_call ) + 3;

	n = read_kmem( &p_sys_call_table, p_system_call+offset, 4, 
		file_descriptor_kmem );
	if ( n < 0 ) {
		return( n );
	}

	if ( p_sys_call_table != get_symbol_by_name("sys_call_table") ) {
		fatal( "[#] system_call function hijacked ! "
			"new sys_call_table at %#x\n", p_sys_call_table );
		fatal( "[#] trying to work anyway..\n\n" );
		if ( force_flag ) {
			p_sys_call_table = get_symbol_by_name( "sys_call_table" );
			fprintf( stderr, "Trying to force the sys_call_table address using %#08x\n",
				p_sys_call_table );
		} else {
			fatal( "If it fail now try to use the -f flag !\n" );
		}
		return( 1 );
	}
	return( 0 );
}

void
get_list_addr_syscall( void )
{
	char buf[ BUFSIZ ];
	char buf2[ BUFSIZ ];
	int i;

	for ( i = 0; i < NR_syscalls; i+=2 ) {
		snprintf( buf, sizeof(buf), "[%3i] %20s: 0x%08lx", i, list_syscall[i],
			(unsigned long)sys_call_table[i] );
		if ( i + 1 < NR_syscalls ) { /* if NR_syscall is odd */
			snprintf( buf2, sizeof(buf2), "[%3i] %20s: 0x%08lx", i+1, list_syscall[i+1], 
			(unsigned long)sys_call_table[i+1] );
			if ( strlen(buf)+strlen (buf2)+2 <= 80 ) {
				fprintf( stdout, "%s  %s\n", buf, buf2 );
			} else {
				fprintf( stdout, "%s\n%s\n", buf, buf2 );
			}
		} else {
			fprintf( stdout, "%s\n", buf );
		}
	}
	return;
}

int
refresh_syscall_table( void )
{
	int n;
	
	n = read_kmem( &sys_call_table, p_sys_call_table,
		sizeof(sys_call_table), file_descriptor_kmem );
	return( n );
}

void
hijack_syscall( int number, size_t kaddr )
{
	int n;

	n = write_kmem( &kaddr, p_sys_call_table+(sizeof(void *)*number),
		sizeof(kaddr), file_descriptor_kmem );
	if ( n < 0 ) {
		fatal( "writing in %s failed !\n", KMEM );
	}

	/* sys_call_table is modified, so we initialize again 
	 * sys_call_table 
	 */
	n = refresh_syscall_table();
	if ( n < 0 ) {
		return;
	}
	return;
}

/* That is not enough to check the first 7 bits representing a jmp. 
 * I thus prefer to make a file of signatures of the syscalls on a 
 * healthy system and to make a checking starting from this check.
 */
void
signatures_syscall( char * signatures_file, int size_sign )
{
#define CREATE_SYSCALL "signature of syscalls\n"
	create_file_signatures( CREATE_SYSCALL, signatures_file, size_sign,
	SIG_FOR_SYSCALL );
	return;
}

void
check_syscall( char * signatures_file, int size_sign )
{
	int i;
	
	i = file_is_present( signatures_file );
	if ( i < 0 ) {
		return;
	}

#define CHECK_SYSCALL "\nn\tsignature\t\tactual\n" \
	"----------------------------------------------------------\n"
	i = check_file_signatures( CHECK_SYSCALL, signatures_file, size_sign,
		SIG_FOR_SYSCALL );
	if ( i == EXIT_SUCCESS ) {
		fputs( "syscall_table valid..\n\n", stdout );
	}
	return;
}

static size_t start_address;
static size_t end_address;

void
get_definition_syscall( int nr )
{
	fputs( "         name : ", stdout );
	if ( !strlen(list_syscall[nr]) ) {
		fputs( "(null)\n", stdout );
	}
	else {
		fprintf( stdout, "%s\n", list_syscall[nr] );
	}
	
	fprintf( stdout, "   syscall n : %d\n", nr );
	start_address = (unsigned int)sys_call_table[nr];
	fprintf( stdout, "start address : %#x\n", start_address );

	end_address = get_next_symbol_by_addr( start_address );
	if ( end_address != 0 ) {
		fprintf( stdout, "  end address : %#x\n", end_address );
		fprintf( stdout, "       length : %d\n", 
			end_address-start_address );
		fputs( "\n", stdout );
	}
	return;
}

void
get_definition_syscall_by_number( int nr )
{
	get_definition_syscall( nr );
}

void
get_definition_syscall_by_addr( char * address )
{
	size_t p_syscall;
	int i;
	int n;
	int nr = 0;

	p_syscall = strtoul( address, NULL, 16 );
	if ( p_syscall < PAGE_OFFSET || p_syscall > 0xffffffff ) {
		fatal( "address not valid..\n" );
		return;
	}

	for ( i = 0, n = 0; i < NR_syscalls; i++ ) {
		if ( p_syscall == (unsigned int)sys_call_table[i] ) {
			nr = i;
			n++;
			if ( n > 1 ) {
				fprintf( stdout, "address of syscall "
				"sys_ni_syscall..\n\n" );
				return;
			}
		}
	}
    
	if ( nr ) get_definition_syscall( nr );
	else fatal( "syscall at %#x() don't exist..\n", p_syscall );

	return;
}

void
get_definition_syscall_by_name( char * name_syscall )
{
	int i;
	int nr = 0;

	for ( i = 0; i < NR_syscalls; i++ ) {
		if ( !strcmp(list_syscall[i], name_syscall) ) {
			nr = i;
		}
	}

	if ( nr ) get_definition_syscall( nr );
	else fatal( "syscall %s don't exist..\n\n", name_syscall );

	return;
}

void
get_strange_syscall( void )
{
	int i = 0;
	int positive = 0 ;

	if ( refresh_syscall_table () < 0 ) {
		perror( "get_stange_syscall:refresh_syscall_table()" );
		return;
	}

	for ( i = 0; i < NR_syscalls; i++ ) {
		if ( (unsigned int) sys_call_table[i] < PAGE_OFFSET + 0x00100000
			|| (unsigned int) sys_call_table[i] > PAGE_OFFSET + 0x00200000 ) {
			fprintf( stdout, "WARNING : syscall n %d not in the good place : %#08lx\n",
				i, sys_call_table[i] );
			positive++;
			continue;
		}
	}

	if ( !positive ) {
		fprintf( stdout,"All tests passed !\n");
	}
	return;
}

int
scall( int num, int argc, unsigned int argv[] ) 
{
    int __ret;

    if ( argc > 5 ) {
        fatal( "Too many arguments !\n" );
        return( -1 );
    }

    switch ( argc ) {
        case  0:
            SYSCALL0( num );
            break;
        case 1:
            SYSCALL1( num, argv[0] );
            break;
        case 2:
            SYSCALL2( num, argv[0], argv[1] );
            break;
        case 3:
            SYSCALL3 (num, argv[0], argv[1], argv[2] );
            break;
        case 4:
            SYSCALL4 (num, argv[0], argv[1], argv[2], argv[3] );
            break;
        case 5:
            SYSCALL5 (num, argv[0], argv[1], argv[2], argv[3], argv[4] );
            break;
        default:
            fatal( "Error in the number of arguments !\n" );
            return( -1 );
    }
    return( __ret );
}

void
usage_sc( void )
{
    usage_t p_usage;

    p_usage.command_line =
        "sc [-n] [-imc] [-rs][t] (the order of [-icm] is VERY important !)";
    p_usage.description =
		"sc permits to directly call a syscall with \"well\" \n"
		"formated arguments ...";
    p_usage.arguments = 
		" -n [num] : the syscall # (take a look at sl) ;\n"
		" -i [value] : this argument must be passed like an integer ;\n"
		" -c [value] : this argument must be passed like a string\n"
		" (you could add spaces but don't use '\"') ;\n"
		" -m [size] :  this argument is a malloced part of memory of [size] ;\n"
		" -r : to display the value as an hexadecimal value ;\n"
		" -s : to display the value as a string ;\n"
		" -t [num] : to display the value of element num when returning from sc ;\n"
		"(Need to be preceded by a -r or -s).";

    usage( &p_usage );
    return;
}

enum {
	NOT,
	INT,
	CHA,
};

void
do_sc( int pargc, char * pargv[] )
{
    char c = 0;
    int sys_num = 0;
    int argc = 0;
    unsigned int arg_i[6];
    char * line = NULL;
    char * arg_c[6];
    int arg_r[6];
    int debug_flag = 0;
    int val_ret = 0;
    
    NOARGUMENTS( pargc, 1, do_sc );
	optind = 0;

    memset( arg_i, 0, sizeof(arg_i) );
    memset( arg_c, 0, sizeof(arg_c) );
    memset( arg_r, 0, sizeof(arg_r) );


    while ( (c = getopt(pargc,pargv,"n:c:i:m:drst:")) != EOF ) {
        switch ( c ) {
            case 'd': /* Not Documented */
                debug_flag = 1;
                break;
            case 'r':
                val_ret = INT;
                break;
            case 's':
                val_ret = CHA;
                break;
            case 't':
                if ( atoi(optarg) < 6 && atoi(optarg) > 0 ) {
					arg_r[atoi(optarg)-1] = val_ret;
                }
                break;
            case 'n':
                sys_num = atoi( optarg );
                CHECK_NUMBER_SYSCALL( sys_num );
                break;
            case 'i':
                arg_i[argc++] = strtoul( optarg, NULL, 0 );
                break;
            case 'c':
                line = malloc( strlen(optarg)+1 );
                sprintf( line, "%s", optarg );
                while ( optind < pargc && pargv[optind][0] != '-' ) {
                    int size = strlen( line );
                    line = realloc( line, size+1+strlen(pargv[optind])+1 );
					if ( line == NULL ) {
						perror( "do_sc:realloc()" );
						return;
					}
                    sprintf( line+size, " %s", pargv[optind] );
                    optind++;
                }
                arg_c[argc] = line;
                arg_i[argc++] = (unsigned int)line;
                break;
            case 'm':
                arg_c[argc] = (char *)arg_i[argc] = (unsigned int)malloc( atoi(optarg) );
                bzero( (char*)arg_i[argc], atoi(optarg) );
                argc++;
                break;
            default:
                usage_sc();
                return;
        }
    }
	
    if ( sys_num != -1 ) {
        int i = 0;
		
        if ( debug_flag ) {
            fprintf( stderr, "Get the following Arguments :\n" );
            for ( i = 0; i < argc; i++ ) {
                if ( arg_c[i] == 0 ) {
                    fprintf( stderr, "Integer\t : %#08x\n", arg_i[i] );
                } else {
                    fprintf( stderr, "String\t (\\n before and after) :BEGIN\n%s\nEND\n", 
						(char *)arg_i[i] );
                }
            }
        }
		
        scall( sys_num, argc, arg_i );
        for ( i = 0; i < argc; i++ ) {
            switch ( arg_r[i] ) {
                case INT:
                    fprintf( stdout, "Return arg %i : (Integer) : %u - %#08x\n", 
						i+1, arg_i[i], arg_i[i] );
                    break;
                case CHA:
                    fprintf( stdout, "Return arg %i : (String) :BEGIN\n%s\nEND\n", 
						i+1,(char*)arg_i[i] );
                    break;
                default:
                    break;
            }
			
            if ( arg_c[i] != NULL ) {
                free( arg_c[i] );
                arg_c[i] = NULL;
            }
        }
    } else {
        usage_sc();
    }
    return;
}

void
usage_sr( void )
{
    usage_t p_usage;

    p_usage.command_line = 
		"sr [ 0 | 1 | 2 | 3 ]";
    p_usage.description =
		"This command makes it possible to manually refresh syscall_table and\n"
		"system_call address.";
    p_usage.arguments = 
    	"  0 : syscall_table contents ;\n"
    	"  1 : syscall_table address ;\n"
    	"  2 : system_call address ;\n"
		"  3 : all.";

    usage( &p_usage );
    return;
}

void
do_sr( int pargc, char * pargv[] )
{
	int i;
	
    NOARGUMENTS( pargc, 1, sr );

    optind = 0;

	i = atoi( pargv[1] );
	if ( i < 0 || i > 3 ) {
		usage_sr();
		return;
	}

	switch ( i ) {
		case 0:
			refresh_syscall_table();
			break;
		case 1:
			set_syscall_table();
			break;
		case 2:
			set_system_call();
			break;
		case 3:
			refresh_syscall_table();
			set_syscall_table();
			set_system_call();
			break;
		default :
			break;
	}
	return;
}


void
usage_sl( void )
{
    usage_t p_usage;

    p_usage.command_line = NULL;
    p_usage.description = "This command list syscalls and theirs addresses.";
    p_usage.arguments = NULL;

    usage( &p_usage );
    return;
}

void
do_sl( int pargc, char * pargv[] )
{
	get_list_addr_syscall();
}

void
usage_sh( void )
{
    usage_t p_usage;

    p_usage.command_line = NULL;
    p_usage.description =
		"This command makes it possible to hijack a syscall\n"
		"sh [syscall number] <syscall address>";
    p_usage.arguments = NULL;

    usage( &p_usage );
    return;
}

void
do_sh( int pargc, char * pargv[] )
{
	int i;
	size_t l;

	if ( pargc != 3 ) {
		usage_sh();
		return;
	}

	i = atoi( pargv[1] );
	CHECK_NUMBER_SYSCALL( i );
	XSTRTOUL( pargv[2], l );

	fprintf( stdout, "hijacked syscall : %d\n", i );
	fprintf( stdout, "    with address : %#x\n", l );
	hijack_syscall( i, l );

	return;
}

void
usage_si( void )
{
    usage_t p_usage;

    p_usage.command_line =
		"si [ -gst ] <arguments>";
    p_usage.description =
		"This command check security about syscall_table";
    p_usage.arguments =
    	"  -g [filename] <size_of_signature> : create a file of signatures ;\n"
    	"  -s [filename] <size_of_signature> : check syscall_table integrity ;\n"
    	"  -t : check for strange syscall.";

    usage( &p_usage );
	return;
}

enum {
	CREATE_SIGNATURE_SYSCALL,
	CHECK_SIGNATURE_SYSCALL,
};

void
do_si( int pargc, char * pargv[] )
{
    int size_sign = 10; /* 10 by default is a minimum ! */
    int flag;
    char c;

    NOARGUMENTS( pargc, 1, sd );

    start_address = end_address = 0;
    optind = 0;
    flag = -1;

    while ( (c = getopt(pargc, pargv, "ts:g:")) != EOF ) {
        switch ( c ) {
			case 't':
				get_strange_syscall();
				break;
            case 's':
                flag = CHECK_SIGNATURE_SYSCALL;
                break;
            case 'g':
                flag = CREATE_SIGNATURE_SYSCALL;
                break;
            default:
                usage_si();
                return;
        }
    }

	switch ( flag ) {
        case CHECK_SIGNATURE_SYSCALL:
            if ( pargc < 3 ) {
                usage_idt();
                return;
            }

            if ( pargc == 4 ) {
                size_sign = atoi( pargv[3] );
                if ( size_sign > 64 ) {
                    fatal( "please, choose size < 64..\n" );
                    return;
                }
            }
            check_syscall( pargv[2], size_sign );
            break;

        case CREATE_SIGNATURE_SYSCALL:
            if ( pargc < 3 ) {
                usage_idt();
                return;
            }

            if ( pargc == 4 ) {
                size_sign = atoi( pargv[3] );
                if ( size_sign > 64 ) {
                    fatal( "please, choose size < 64..\n" );
                    return;
                }
            }
            signatures_syscall( pargv[2], size_sign );
            break;

        default:
            return;
    }
    return;
}

void
usage_sd( void )
{
    usage_t p_usage;

    p_usage.command_line = 
		"sd [ -dan ] <arguments>";
    p_usage.description =
		"This command makes it possible to display characteristics about syscalls.";
    p_usage.arguments =
    	"  -d [syscall number] <opcode|disass> : display characteristics\n"
    	"of a syscall according to a number ;\n"
    	"  -a [address] <opcode|disass> : display characteristics of a\n"
    	"syscall according to an address ;\n"
    	"  -n <syscall name> <opcode|disass> : display characteristics of\n"
    	"a syscall according to a name.";

    usage( &p_usage );
    return;
}

void
do_sd( int pargc, char * pargv[] )
{
    char c;
	int flag;
    int i;

    NOARGUMENTS( pargc, 1, sd );

	start_address = end_address = 0;
    optind = 0;
    flag = -1;

	while ( (c = getopt(pargc, pargv, "d:a:n:")) != EOF ) {
		switch ( c ) {
			case 'd':
				i = atoi( optarg );
				CHECK_NUMBER_SYSCALL( i );
				get_definition_syscall_by_number( i );
				break;
			case 'a':
				get_definition_syscall_by_addr( optarg );
				break;
			case 'n':
				get_definition_syscall_by_name( optarg );
				break;
			default:
				usage_sd();
				return;
		}
	}

    if ( pargv[3] && start_address != 0 && end_address != 0 ) {
        if ( !strcmp(pargv[3], "opcode") ) {
        fprintf( stdout, "[!] opcode from address %#x..\n",
            start_address );
        kdump( start_address, end_address-start_address,
            file_descriptor_kmem );
        }
        if ( !strcmp(pargv[3], "disass") ) {
        fprintf( stdout, "[!] disassembling from address %#x..\n\n",
            start_address );
        kdisass( start_address, end_address-start_address,
            file_descriptor_kmem );
        fputs( "\n", stdout );
        }
    }
	return;
}

void
usage_syscall( void )
{
    usage_t p_usage;

    p_usage.command_line = NULL;
    p_usage.description =
		"This command makes it possible to play with the syscalls:\n"
		"list syscalls and their addresses, syscall hijack..";
    p_usage.arguments =
		"sd : display characteristics of a syscall according to a number ;\n"
		"si : security on syscall_table ;\n"
		"sh : hijack a syscall ;\n"
		"sl : list syscalls and theirs adresses ;\n"
		"sr : manually refresh of syscall_table and system_call address ;\n"
		"sc : call a syscall.\n\n"
		"Type \"help\" followed by command name for full documentation.";

    usage( &p_usage );
	return;
}
