/*
 * The project kernsh is (C) 2002-2004. This program is free software;
 * you can redistribute it and/or modify it under the terms of the GNU
 * General Public License as published by the Free Software Foundation;
 * Version 2. This guarantees your right to use, modify, and
 * redistribute this software under certain conditions.
 *
 * Source is provided to this software because we believe users have a
 * right to know exactly what a program is going to do before they run
 * it.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details (
 * http://www.gnu.org/copyleft/gpl.html ).
 *
 */

#define __KERNEL__
#ifdef GCC33
#define __ASM_SYSTEM_H
#define local_irq_save(x) do { x=x; } while (0);
#define local_irq_enable(x) do { } while (0);
#define local_irq_disable(x) do { } while (0);
#define local_irq_restore(x) do { } while (0);
#define cmpxchg(x,y,z) 1;
#define xchg(x,y) y;
#define smp_mb() return ;
#endif
#include <linux/mm.h>
#include <net/sock.h>
#undef  __KERNEL__
#include "include/kernsh.h"

extern void *malloc (size_t );
extern void free(void *);

extern int file_descriptor_kmem;

/* the symbol of a global task_struct structure */
#define TASK_STRUCT_SYM "keventd_task"
#define TASK_UNION_SYM "init_task_union"

size_t 
find_task_struct( void )
{
	size_t addr;

	if ( !(addr = get_symbol_by_name(TASK_UNION_SYM)) ) {
		if ( !(addr = get_symbol_by_name(TASK_STRUCT_SYM)) ) {
			fatal( "unable to retrieve symbol \"%s\", exiting.\n", 
				TASK_STRUCT_SYM );
			return( 0 );
		}
	}
	return( addr );
}

/*
 * Using get_next_struct :
 * 1- You first call get_next_struct (0), so its initialize
 * 2- You then call it until you receive NULL, and you will
 * get all task_struct available.
 *
 * XXX : possible race condition
 */
int 
get_next_struct( struct task_struct * ret )
{
	static size_t ts_addr = 0;
	static size_t ts_start = 0;
	size_t addr_ret = 0;
	int i;
    
	if ( !ts_addr || !ret ) {
		ts_addr = find_task_struct();
		if ( !ret ) {
			ts_start = 0;
			return( 1 );
		}
	}

	if ( ts_addr == ts_start ) {
		ts_addr = ts_start = 0;
		return ( 0 );
	}
	
	if ( !ts_start ) {
		ts_start = ts_addr;
	}

	i = read_kmem( ret, ts_addr, sizeof(struct task_struct), 
		file_descriptor_kmem ); 
	if ( i == -1 ) {
		fatal( "error : unable to read task_struct in mem\n" );
		return( 0 );
	}

	addr_ret = ts_addr;
#ifdef  KERNEL_22
    ts_addr = (size_t)ret->next;
#elif RED_HAT_LINUX_KERNEL && defined (next_task)
	/* This is for fedora kernel Hope that it won't break too many things*/
	ts_addr = (size_t)list_entry(ret->tasks.next, struct task_struct, tasks);
#else
    ts_addr = (size_t)ret->next_task;
#endif
	return( addr_ret );
}

struct task_struct * 
get_process_by_pid( pid_t pid )
{
	static struct task_struct ret;

	if ( !get_next_struct(NULL) ) {
		return( NULL );
	}

    while ( get_next_struct(&ret) ) {
		if ( ret.pid == pid ) {
			return( &ret );
		}
	}

    return( NULL );
}

struct task_struct * 
get_process_by_name( char * name )
{
	static struct task_struct ret;

    if ( !get_next_struct(NULL) ) {
		return( NULL );
	}
	
	while ( get_next_struct(&ret) ) {
		if ( !strcmp(ret.comm, name) ) {
			return( &ret );
		}
	}
	return( NULL );
}

void 
proc_kps( void )
{
	struct task_struct ret;

	if ( !get_next_struct(NULL) ) {
		return;
	}

	fputs( "PID\t\tCOMMAND\n", stdout );
    while ( get_next_struct(&ret) ) {
		fprintf( stdout, "%d\t\t%s\n", ret.pid, ret.comm );
	}
    return;
}

void 
proc_khide_detect( void )
{
	struct task_struct ret;
	char c;
	int addr = 0;
	int pwet = 0;
	int hidded = 0;
	unsigned int pid = 0;

    fputs( "Detecting hiden process .", stdout );
    fflush( stdout );

	if ( !get_next_struct(NULL) ) {
		return;
	}

	/* The following is here to get ride of the swapper processus
	 * which display falsy as an hidden processus
	 */
	addr = get_next_struct( &ret );
	/* Then launch our dogs :) */
	while ( (addr = get_next_struct(&ret)) ) {
		pwet = 0;
		if ( is_hided(ret.pid) ) {
			if ( !pwet ) {
				fputs( "\n", stdout );
			}
			hidded++;
			pwet++;
			fprintf ( stdout, "Pid %d (%s) is\t"VERT"HIDED"NORM" !\n", 
				ret.pid, ret.comm );
			fprintf( stdout, "Uid : %d Gid : %d\n", ret.uid, ret.gid );
			if ( ret.pid == 0 ) {
				fprintf( stdout, "Unhided it ?" );
				fflush( stdout );
				c = getchar();
				fflush( stdin );
				if ( c == 'y' || c == 'Y' ) {
					getchar();
					fprintf( stdout, "Please enter the pid you want to put here\n"
						"(MUST to be not used by an other process !): " );
					fflush( stdout );
					if ( 1 == fscanf(stdin,"%u",&pid) ) {
						ret.pid = (pid_t)pid;
						write_kmem( &ret, addr, sizeof(ret), file_descriptor_kmem );
						fputs( "Done\n", stdout );
					} else {
						fputs( "Invalid Pid !\n", stdout );
					}
				} else {
					getchar();
				}
			}
		}
		if ( !pwet ) {
			fputs( ".", stdout );
			fflush( stdout );
		}
	}

	if ( !hidded ) {
		fputs( "\nNo hiden process were detected, you look safe :)\n", 
			stdout );
	}
    return;
}

void 
proc_khide( pid_t pid )
{
	struct task_struct ret;
	size_t addr;
	int duh = 0;
	char c;

	fprintf( stdout, "Looking for processus %d\t", pid );
	fflush( stdout );

	if ( !get_next_struct(NULL) ) {
		return;
	}

	while ( (addr=get_next_struct(&ret)) ) {
		if ( ret.pid == pid ) {
			ret.pid = 0;
			fputs( VERT"Ok\n\n"NORM, stdout );
			fprintf( stdout, "struture at %#08x\n", addr );
			fprintf( stdout, "Name : %s | Uid : %d | Gid : %d\n", 
				ret.comm, ret.uid, ret.gid );
			fputs( "WARNING !! This operation can crash the kernel :\n"
				"If the hided processus exit, init will be kill,\n"
				"crashing the kernel in the same time. Continue ? y/N\t", 
				stdout );
			fflush( stdout );
			c = getchar ();
			getchar ();
			if ( c == 'y' || c == 'Y' ) {
				ret.pid = 0;
				write_kmem( &ret, addr, sizeof(ret), 
					file_descriptor_kmem );
				fputs( "Done\n", stdout );
			} else {
				fputs( "Abording !\n", stdout );
			}
			duh++;
			break;
		}
	}

    if ( !duh ) fputs( "Not Found\n", stdout );
    return;
}

void 
proc_kts( pid_t pid )
{
	/* Well, this is very hard to implement because there is a lot's of 
	 * differences depending on the distro you use ...
	 */
	fputs( "Not implemented ..\n", stdout );
    return;
}

#define MAX_DIR_DEPTH 1024

char * socket_state_char[] = { 
	"Free",
	"Unconnected",
	"Connecting",
	"Connected",
	"Disconnecting"
};

void
print_inode_info( struct inode ino )
{
	if (S_ISFIFO( ino.i_mode )) {
		struct pipe_inode_info ip;
		fprintf(stdout,"pipe # : %lu ", ino.i_ino);
		if ( read_kmem((char*)&ip,(unsigned int)ino.i_pipe,sizeof(ip),file_descriptor_kmem) < 0) {
			return;
		}
#if KERNELVERSION >= 0x020400
		fprintf(stdout,"(base : %p, len : %d, "
				"writers : %d, readers : %d)\n", ip.base,
				ip.len, ip.writers,ip.readers);
#else
		fprintf(stdout,"(base : %p, "
				"writers : %d, readers : %d)\n", ip.base,
				ip.writers,ip.readers);
#endif
/*		hexdump((char*)&ip,sizeof(ip),0);*/
	} else if (S_ISSOCK( ino.i_mode )) {
		struct proto_ops ops;
		fprintf(stdout,"state : %s "/*, type : %d, flags : %lx, Sock : %x\n"*/,
				socket_state_char[ino.u.socket_i.state]/*,ino.u.socket_i.type,ino.u.socket_i.flags,(unsigned int)ino.i_sock*/);
		if ( read_kmem((char*)&ops,(unsigned int)ino.u.socket_i.ops,sizeof(ino.u.socket_i.ops),file_descriptor_kmem) < 0) {
			return;
		}
		switch(ops.family) {
				struct sock sk; /* for gcc-2.95 :/ */
			case AF_UNIX:
				fprintf(stdout,"(Unix)");
				break;
			case AF_INET:
				fprintf(stdout,"(Net)");
				read_kmem((char *)&sk,(unsigned int)ino.u.socket_i.sk,sizeof(sk),file_descriptor_kmem);
				fprintf(stdout," %u.%u.%u.%u:%u ->",(unsigned int)((unsigned char*)&sk.rcv_saddr)[0],(unsigned int)((unsigned char *)&sk.rcv_saddr)[1],(unsigned int)((unsigned char*)&sk.rcv_saddr)[2],(unsigned int)((unsigned char *)&sk.rcv_saddr)[3],ntohs(sk.sport));
//				read_kmem((char *)&addr,(unsigned int)ino.u.socket_i.sk+4,sizeof(addr),file_descriptor_kmem);
				fprintf(stdout," %u.%u.%u.%u:%u",(unsigned int)((unsigned char*)&sk.daddr)[0],(unsigned int)((unsigned char *)&sk.daddr)[1],(unsigned int)((unsigned char*)&sk.daddr)[2],(unsigned int)((unsigned char *)&sk.daddr)[3],ntohs(sk.dport));
				break;
			default :
				fprintf(stdout,"(Unknow : %x)",ops.family);
				break;
		}
		fprintf(stdout," \n");
/*		hexdump((char*)&ino.u.socket_i,sizeof(struct socket),0);*/
		/* Note : There is still a lot's of work to do :/ */
	} else {
		fprintf(stdout,"inode # : %lu uid : %d gid : %d, size : %lld, rdev : %x\n",
				ino.i_ino,(int)ino.i_uid,(int)ino.i_gid,ino.i_size,ino.i_rdev);
	}
}

void
print_file_info(struct file tmpfile)
{
	char * dirnames[MAX_DIR_DEPTH];
	struct dentry tmpdir,partmpdir, *ptmpdir = NULL;
	struct inode tmpino,partmpino;
	int n =0,j=0,skip=0;
	
	n = read_kmem((char*)&tmpdir,(unsigned int)tmpfile.f_dentry,sizeof(tmpdir),file_descriptor_kmem);
	if ( n < 0 ) {
		fprintf(stdout,"Ok\n");
		return;
	}
	
	ptmpdir = &tmpdir ;
	while( j < MAX_DIR_DEPTH ) {
		dirnames[j] = malloc(tmpdir.d_name.len + 1);
		memset(dirnames[j],0,tmpdir.d_name.len + 1 );
		n = read_kmem(dirnames[j],(unsigned int)tmpdir.d_name.name,tmpdir.d_name.len,file_descriptor_kmem);
		if ( n < 0 ) { return ; }
		j++;
		if (tmpdir.d_parent != NULL && tmpdir.d_inode != NULL ) {
			n = read_kmem((char*)&partmpdir,(unsigned int)tmpdir.d_parent,sizeof(partmpdir),file_descriptor_kmem);
			if ( n < 0 ) {
				break;
			}
			if ( partmpdir.d_inode != NULL ) {
				n = read_kmem((char*)&tmpino,(unsigned int)tmpdir.d_inode,sizeof(tmpino),file_descriptor_kmem);
				if ( n < 0 ) {
					break;
				}
				n = read_kmem((char*)&partmpino,(unsigned int)partmpdir.d_inode,sizeof(partmpino),file_descriptor_kmem);
				if ( n < 0 ) {
					break;
				}
				if (tmpino.i_ino == partmpino.i_ino ) {
					break;
				}
			}
		}
		ptmpdir = tmpdir.d_parent;
		if ( ptmpdir == NULL ) { break; }
		n = read_kmem((char*)&tmpdir,(unsigned int)tmpdir.d_parent,sizeof(tmpdir),file_descriptor_kmem);
		if ( n < 0 ) {
			break;
		}
	}
	for( j = j-1; j>=0; j--) {
		if (dirnames[j][0] != '/' && skip == 0) {
			fprintf(stdout,"/");
		} else if ( dirnames[j][0] != '/' && skip != 0 ) {
			/* XXX fix the number of '/' */
			skip = 0;
		} else {
			skip = 1;
		}
		fprintf(stdout,"%s",dirnames[j]);
		free(dirnames[j]);
	}
	fprintf(stdout," ");
	n = read_kmem((char*)&tmpdir,(unsigned int)tmpfile.f_dentry,sizeof(tmpdir),file_descriptor_kmem);
	if ( n < 0 ) {
		return;
	}

	if (tmpdir.d_inode != NULL) {
		n = read_kmem((char*)&tmpino,(unsigned int)tmpdir.d_inode,sizeof(tmpino),file_descriptor_kmem);
		if (n < 0) {
			return;
		}
		print_inode_info(tmpino);
	}
}

void 
proc_kfile( pid_t pid )
{
	struct task_struct *ret = get_process_by_pid( pid );
	if ( ret == NULL ) {
		fatal("Process not found !\n");
		return;
	}
	if ( ret->files != NULL ) {
		struct file * stmpfile[OPEN_MAX];
		struct file tmpfile /* = NULL*/;
		struct files_struct fs;
		int i,j;
		int n = 0;
		j = 0;
		n = read_kmem((char*)&fs,(unsigned int)ret->files,sizeof(fs),file_descriptor_kmem);
		if ( n < 0 ) {
			fatal("proc_kfile : read_kmem failed");
			return ;
		}

		fputs( "\n[ files_struct ]\n", stdout );
		fprintf( stdout, "count              : %d\n", (int)fs.count.counter );
		fprintf( stdout, "file_lock          : %p\n", (void*)0/*fs.file_lock*/ );
		fprintf( stdout, "max_fds            : %d\n", fs.max_fds );
		fprintf( stdout, "max_fdset          : %d\n", fs.max_fdset );
		fprintf( stdout, "next_fd            : %d\n", fs.next_fd );
		fprintf( stdout, "fd (array)         : %p\n", fs.fd );
		fprintf( stdout, "close_on_exec      : %p\n", fs.close_on_exec );
		fprintf( stdout, "open_fds           : %p\n", fs.open_fds );
		fprintf( stdout, "close_on_exec_init : %p\n", fs.close_on_exec_init.fds_bits );
		fprintf( stdout, "open_fds_init      : %p\n", fs.open_fds_init.fds_bits );
		fprintf( stdout, "fd_array           : %p\n\n", fs.fd_array );

		n = read_kmem((char*)&stmpfile,(unsigned int)fs.fd,sizeof(stmpfile),file_descriptor_kmem);
		if ( n < 0 ) {
			fatal("proc_kfile : read_kmem failed");
			return ;
		}
		
		fprintf(stdout,"[ file descriptor ]\n");
		for( i = 0; i < fs.max_fds /*OPEN_MAX*//*stmpfile[i] != NULL*/; i++ ) {
			if ( stmpfile[i] == NULL ) {
				continue;
			}
			fprintf(stdout," %d :\t",i);
			n = read_kmem((char*)&tmpfile,(unsigned int)stmpfile[i],sizeof(tmpfile),file_descriptor_kmem);
			if ( n < 0 ) {
				break;
			}
			print_file_info(tmpfile); /* Print the info */
			j = 0;
		}
	}
	/* Now we need to treat the memory mapped files :/ */
	if( ret->mm != NULL ) {
		struct mm_struct mm;
		int n;
		fprintf(stdout,"\n[ mapped files ]\n");
//		fprintf(stdout,"mm : %p\n",ret->mm);
		n = read_kmem((char*)&mm,(unsigned int)ret->mm,sizeof(mm),file_descriptor_kmem);
		if (n < 0) {
			return;
		}
//		hexdump((char*)&mm,sizeof(mm),0);
		if ( mm.mmap != NULL ) {
			struct vm_area_struct mmap;
			struct file fp;
//			fprintf(stdout,"vm : %p (mmap #): %d \n",mm.mmap,mm.map_count);
			n = read_kmem((char*)&mmap,(unsigned int)mm.mmap,sizeof(mmap),file_descriptor_kmem);
			if (n < 0) {
				return;
			}
//			hexdump((char*)&mmap,sizeof(mmap),0);
			for ( n = 0; n < mm.map_count; n++ ) {
//				fprintf(stdout,"file is : %p\n",mmap.vm_file);
				if ( mmap.vm_file == NULL )	{
					goto next;
				}
				n = read_kmem((char*)&fp,(unsigned int)mmap.vm_file,sizeof(fp),file_descriptor_kmem);
				if (n < 0) {
					return;
				}
				fprintf(stdout," %p to %p:\n",(void*)mmap.vm_start,(void*)mmap.vm_end);
//				hexdump((char*)&fp,sizeof(fp),0);
				print_file_info(fp);
//				fprintf(stdout,"\n");
next:
				if( mmap.vm_next != NULL)
				{
					n = read_kmem((char*)&mmap,(unsigned int)mmap.vm_next,sizeof(mmap),file_descriptor_kmem);
					if (n < 0) {
						return;
					}
				} else {
					break;
				}
			}
		}
	}
    return;
}

void 
proc_kinfo( pid_t pid )
{
	fputs( "Not implemented .. What do you want to see ?\n", stdout );
    return;
}

